/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "aps/Button",
    "aps/Container",
    "aps/load",
    "aps/Memory",
    "aps/Output",
    "aps/xhr",
    "app/tags",
    "app/common",
    "app/messages",
    "app/wizardData",
    "dijit/registry",
    "dojo/promise/all"
], function (
    Button,
    Container,
    load,
    Memory,
    Output,
    xhr,
    tags,
    common,
    messages,
    wizardData,
    registry,
    all
) {
    var tenant = aps.context.vars.tenant,
        administratorNewViewId = common.getRelatedViewId('new'),
        administratorViewViewId = common.getRelatedViewId('view'),
        administratorListViewId = common.getRelatedViewId('list');

    function getUserStatus(status) {
        var content;
        switch (status) {
            case -1:
                content = '<img src=images/user_statuses/ico_state_off_16px.png> ' + _("Not activated");
                break;
            case 0:
                content = '<img src=images/user_statuses/ico_state_off_16px.png> ' + _("Disabled");
                break;
            case 1:
                content = '<img src=images/user_statuses/ico_state_on_16px.png> ' + _("Enabled");
                break;
        }
        return content;
    }

    function onAddNew() {
        aps.apsc.gotoView(administratorNewViewId);
    }

    function onRemove() {
        var self = this;
        if (!confirm(_("Are you sure that you want to remove the administrator account?"))) {
            self.cancel();
            return;
        }

        var grid = registry.byId("adminGrid"),
            selectionArray = grid.get("selectionArray"),
            defers = [],
            deletedIds = [];

        messages.removeAll();

        selectionArray.forEach(function (id) {
            var admin = grid.store.get(id),
                deffer = xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + admin.id, {
                    method: "DELETE",
                    query: {
                        datacenterId: ''
                    }
                }).then(function () {
                    deletedIds.push(id);
                }, messages.displayErrorAndGo);
            defers.push(deffer);
        });

        all(defers).then(function () {
            deletedIds.forEach(function (id) {
                grid.store.remove(id);
                selectionArray.splice(selectionArray.indexOf(id), 1);
            });
            grid.refresh();
            self.cancel();
        });
    }

    function getUserActions(row) {
        var widget;
        switch (row.status) {
            case +common.STATUS.INACTIVE:
                widget = new Button({
                    iconClass: "sb-export",
                    label: _("Resend activation email"),
                    onClick: function () {
                        var self = this;
                        messages.removeAll();
                        xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + row.id + "/resend", {
                            method: "PUT",
                            query: {datacenterId: ''}
                        }).then(
                            function () {
                                messages.displayMessage(_("The activation message has been successfully sent to {0}.").replace("{0}", row.email), 'info');
                                self.cancel();
                            },
                            function () {
                                messages.displayMessage(_("Failed to send the activation message to {0}.").replace("{0}", row.login), 'error');
                                self.cancel();
                            }
                        );
                    }
                });
                break;
            case +common.STATUS.DISABLED:
                widget = new Button({
                    iconClass: "sb-export",
                    label: _("Enable"),
                    onClick: function () {
                        var self = this;
                        messages.removeAll();
                        xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + row.id + "/enable", {
                            method: "PUT",
                            query: {datacenterId: ''}
                        }).then(
                            function () {
                                messages.displayMessage(_("Administrator {0} has been successfully enabled.").replace("{0}", row.login), 'info');
                                xhr.get("/aps/2/resources/" + tenant.aps.id + "/administrators").then(function (administrators) {
                                    var grid = registry.byId("adminGrid");
                                    grid.store.setData(administrators);
                                    grid.refresh();
                                }, messages.displayError);
                            },
                            function () {
                                messages.displayMessage(_("Failed to enable adminisrtator {0}.").replace("{0}", row.login), 'error');
                                self.cancel();
                            }
                        );
                    }
                });
                break;
            case +common.STATUS.ENABLED:
                widget = new Container();
                widget.addChild(new Button({
                    iconClass: "sb-export",
                    label: _("Disable"),
                    onClick: function () {
                        var self = this;
                        messages.removeAll();
                        xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + row.id + "/disable", {
                            method: "PUT",
                            query: {datacenterId: ''}
                        }).then(
                            function () {
                                messages.displayMessage(_("Administrator {0} has been successfully disabled.").replace("{0}", row.login), 'info');
                                xhr.get("/aps/2/resources/" + tenant.aps.id + "/administrators").then(function (administrators) {
                                    var grid = registry.byId("adminGrid");
                                    grid.store.setData(administrators);
                                    grid.refresh();
                                }, messages.displayError);
                            },
                            function () {
                                messages.displayMessage(_("Failed to disable adminisrtator {0}.").replace("{0}", row.login), 'error');
                                self.cancel();
                            }
                        );
                    }
                }));
                break;
        }
        return widget;
    }

    all({
        administrators: xhr.get("/aps/2/resources/" + tenant.aps.id + "/administrators"),
        administratorGuideLink: xhr.get("/aps/2/resources/" + tenant.aps.id + "/administratorGuideLink")
    }).then(function (response) {
        var administrators = response.administrators;

        function getOutput(label, row) {
            return new Output({
                content: '<a href=#>${value}</a>',
                value: label,
                escapeHTML: false,
                onClick: function () {
                    wizardData.setWizardData({
                        adminId: row.id
                    }, administratorViewViewId);
                    aps.apsc.gotoView(administratorViewViewId);
                }
            });
        }

        var columns = [
            {
                field: "id", name: "ID", visible: false
            },
            {
                field: "login", name: _("Login"),
                renderCell: function (row) {
                    return getOutput(row.login, row);
                }
            },
            {
                field: "email", name: _("Email"),
                renderCell: function (row) {
                    return getOutput(row.email, row);
                }
            },
            {
                field: "status", name: _("Status"), escapeHTML: false,
                renderCell: function (row, data) {
                    return getUserStatus(data);
                }
            },
            {
                field: "status", name: "", sortable: false,
                'class': "text-right",
                renderCell: function (row) {
                    return getUserActions(row);
                }
            }
        ];

        load(["aps/PageContainer", {id: "apsPageContainer", "class": "acronis-administrator"}, [
            ["aps/Container", {"id": common.genId('container'), title: _("Information")}, [
                ['aps/Output', {
                    value: _("This is a unique integration of backup with full-stack next-generation anti-malware protection and comprehensive endpoint management tools. This synergy eliminates complexity, so service providers can protect customers better while keeping costs down.") +
                    tags.br() + _("Log into Management Console as an administrator to manage your customers and their backups.") +
                    tags.br() + _("Learn more about the service - Administrator's guide.") +
                    tags.br() + tags.br() + tags.a(response.administratorGuideLink, _("Administrator's guide"), {'target' : '_blank'}) +
                    tags.br() + tags.br(),
                    escapeHTML: false
                }],
                ["aps/Button", {
                    title: _("Log into Management Console"),
                    autoBusy: false,
                    onClick: function () {
                        var self = this,
                            w = window.open('', '_blank');
                        xhr.get("/aps/2/resources/" + tenant.aps.id + "/redirectUrl").then(function (response) {
                            w.location.href = response.url;
                        }, function (error) {
                            w.close();
                            console.error("ERR: \n" + JSON.stringify(error));
                            messages.displayError(error);
                            self.cancel();
                        });
                    }
                }],
                ['aps/Output', {
                    value: "<br><br>",
                    escapeHTML: false
                }]
            ]],
            ["aps/Container", {"id": common.genId('container'), title: _("Administrators")}, [
                ["aps/Grid", {
                    id: "adminGrid",
                    selectionMode: "multiple",
                    apsResourceViewId: administratorViewViewId,
                    showPaging: false,
                    pageSizeOptions: "GRID_UNLIMITED",
                    columns: columns,
                    store: new Memory({data: administrators, idProperty: "id"})
                }, [
                    ["aps/Toolbar", {id: "adminToolbar"}, [
                        ["aps/ToolbarButton", {
                            id: "adminNew",
                            iconClass: "sb-add-new",
                            label: _("Add"),
                            onClick: onAddNew
                        }],
                        ["aps/ToolbarSeparator"],
                        ["aps/ToolbarButton", {
                            id: "adminDelete",
                            iconClass: "sb-delete",
                            label: _("Remove"),
                            requireItems: true,
                            onClick: onRemove
                        }]
                    ]]
                ]]
            ]]
        ]]).then(function () {
            messages.removeAll();
            var data = wizardData.getWizardData(administratorListViewId);
            if (data && data.message) {
                messages.displayMessage(data.message.text, data.message.type);
                wizardData.resetWizardData(administratorListViewId);
            }
        });
    }, messages.displayError);
});
