/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "app/common",
    "app/quotas",
    "app/messages",
    "aps/ResourceStore",
    "aps/WizardData",
    "aps/load",
    "aps/xhr",
    "dijit/registry",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "aps/Memory",
    "dojo/when",
    "dojo/promise/all"
], function (
    common,
    quotas,
    messages,
    Store,
    WizardData,
    load,
    xhr,
    registry,
    at,
    getStateful,
    getPlainValue,
    Memory,
    when,
    all
) {
    var tenant = aps.context.vars.tenant,
        tenantId = tenant.aps.id,
        userData = WizardData.get(),
        user = aps.context.params.user,
        login = user.login.toLowerCase(),
        maxQuotaSessionKey = 'setMaxQuota_' + login,
        setMaxQuota = !userData || !userData[login] || JSON.parse(window.sessionStorage.getItem(maxQuotaSessionKey)),
        subscriptionStore = new Store({
            target: "/aps/2/resources/" + tenantId + "/subscription",
            idProperty: "aps.id"
        }),
        controls = null;

    window.sessionStorage.setItem(maxQuotaSessionKey, setMaxQuota.toString());
    function renderQuota(quotasStatistic, setMaxQuota, userQuota) {
        var quotaContainer = registry.byId("quotaContainer"),
            offeringItemsToDisplay = common.getAvailableCountersForArea('quota_controls_ccpv1_new_user');
        controls = quotas.getQuotaControlsForNewUser(tenant, offeringItemsToDisplay, quotasStatistic, setMaxQuota, userQuota);

        Object.keys(controls).forEach(function (controlKey) {
            quotaContainer.addChild(controls[controlKey]);
        });
    }

    function setAdminPrivilegesChecked(value) {
        registry.byId("adminPrivileges").set("checked", !!value);
    }

    function restoreUserData(userData, userQuota) {
        setAdminPrivilegesChecked(userData.privileged);

        applyQuotaValues(userData, userQuota);
    }

    function applyQuotaValues(userData, userQuota) {
        let getLimitValue = function (limitFieldValue, isStorage) {
            return limitFieldValue === "" ? common.QUOTA_UNLIMITED_VALUE
                : limitFieldValue * (isStorage ? common.BYTES_IN_GB : 1);
        }
        if (!controls) {
            return;
        }
        Object.entries(controls).forEach(function ([counterName, control]) {
            let offeringItem = common.getUserOfferingItem(userData, counterName, true);
            if (offeringItem === null) {
                return;
            }
            offeringItem.limit = control.value
                ? getLimitValue(control.value, common.isStorage(counterName))
                : userQuota === common.USER_QUOTA.STRICT ? 0 : -1;
        });
    }

    all({
        subscriptions: subscriptionStore.query(),
        profiles: xhr.get("/aps/2/resources/" + tenantId + "/profile"),
        quotasStatistics: xhr.get("/aps/2/resources/" + tenantId + "/quotasStatistics"),
        listWithAnyExistingUser: xhr.get("/aps/2/resources/" + tenantId + "/users?limit(0,1)")
    }).then(function (response) {
        var subscriptions = response.subscriptions,
            subscription = subscriptions[0],
            backupProfile = response.profiles.length ? response.profiles[0] : null,
            quotasStatistics = response.quotasStatistics,
            userQuota = backupProfile ? backupProfile.userQuota : common.USER_QUOTA.STRICT,
            isFirstUser = !response.listWithAnyExistingUser.length,
            userNotifications = backupProfile && backupProfile.notifications ? backupProfile.notifications : [],
            notifications = common.transformNotificationsToObject(
                userData && userData[login] ? userData[login].notifications : userNotifications);
        var model = {
            aps: {type: common.TYPES.USERS},
            tenant: [
                {aps: {id: tenantId}}
            ],
            privileged: isFirstUser ? common.USER_PRIVILEGED.YES : common.USER_PRIVILEGED.NO,
            activateDuringCreation: true,
            offeringItems: [],
            notifications: notifications
        };

        load(["aps/PageContainer", {id: "apsPageContainer"}, [
            ["aps/FieldSet", {title: _("Main")}, [
                ["aps/Output", {
                    label: _("User name"),
                    value: common.escapeAngularBrackets(user.displayName + ' (' + user.login + ')')
                }],
                ["aps/CheckBox", {
                    id: "adminPrivileges",
                    label: _("Enable administrator privileges"),
                    checked: at(model, "privileged").transform({
                        format: function (value) {
                            return value === common.USER_PRIVILEGED.YES;
                        },
                        parse: function (value) {
                            return (value) ? common.USER_PRIVILEGED.YES : common.USER_PRIVILEGED.NO;
                        }
                    }),
                    hint: _('An administrator can manage services of other users and back up Office 365 seats.')
                }]
            ]],
            ["aps/FieldSet", {title: _('Notifications')}, [
                ["aps/FieldSet", {
                    label: "MANAGEMENT PORTAL",
                    title: false
                }],
                ["aps/CheckBox", {
                    label: _('Quota overuse notifications'),
                    checked: at(model.notifications, "quota")
                }],
                ["aps/CheckBox", {
                    label: _('Scheduled usage reports'),
                    checked: at(model.notifications, "reports")
                }],
                ["aps/FieldSet", {
                    label: "CYBER PROTECTION",
                    title: false
                }],
                ["aps/CheckBox", {
                    label: _('Failure notifications'),
                    checked: at(model.notifications, "backup_error")
                }],
                ["aps/CheckBox", {
                    label: _('Warning notifications'),
                    checked: at(model.notifications, "backup_warning")
                }],
                ["aps/CheckBox", {
                    label: _('Success notifications'),
                    checked: at(model.notifications, "backup_info")
                }],
                ["aps/CheckBox", {
                    label: _('Daily recap about active alerts'),
                    checked: at(model.notifications, "backup_daily_report")
                }]
            ]],
            ["aps/FieldSet", {
                id: "quotaContainer",
                visible: userQuota != common.USER_QUOTA.OFF,
                description: quotas.getQuotaDescription(userQuota),
                title: _('Quotas')
            }]
        ]]).then(function () {

            if (userQuota != common.USER_QUOTA.OFF) {
                var quotaContainer = registry.byId("quotaContainer");
                quotaContainer.set("description", quotas.getQuotaDescription(userQuota));
                quotaContainer.set("visible", true);
                renderQuota(quotasStatistics, setMaxQuota, userQuota);
            }

            if (userData && userData[login]) {
                restoreUserData(userData[login], userQuota);
            }
        });

        aps.app.onPrev = function () {
            var dataForPrevStep = {};
            if (Object.keys(model.notifications).length) {
                model.notifications = common.transformNotificationsToArray(model.notifications);
            } else {
                delete model.notifications;
            }
            dataForPrevStep[login] = model;
            WizardData.put(dataForPrevStep);

            aps.apsc.prev();
        };

        aps.app.onNext = function () {
            var page = registry.byId("apsPageContainer"),
                resourcesStore = new Store({
                    target: "/aps/2/resources/" + subscription.aps.id + "/resources",
                    idProperty: "aps.id"
                });

            if (!page.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }

            if (Object.keys(model.notifications).length) {
                model.notifications = common.transformNotificationsToArray(model.notifications);
            } else {
                delete model.notifications;
            }

            applyQuotaValues(model, userQuota);
            xhr.get("/aps/2/resources/" + tenant.aps.id + "/loginExists", {
                query: {
                    login: user.login
                }
            }).then(function (existing) {
                if (existing) {
                    messages.removeAll();
                    messages.displayWarning(_("The account with email address already exists in the Cyber Protection service. Delete the old account or specify a different service user."));
                    aps.apsc.cancelProcessing();
                    return;
                }
                resourcesStore.query().then(function (resources) {
                    for (var i = 0; i < resources.length; i++) {
                        if (resources[i].apsType == common.TYPES.USERS) {
                            if (resources[i].usage >= resources[i].limit) {
                                messages.displayWarning(_("Limit for service users has been reached."));

                                aps.apsc.cancelProcessing();
                                return;
                            }
                        }
                    }
                    aps.apsc.gotoView("fake", null, {
                        objects: [getPlainValue(model)],
                        userAttr: "user"
                    });
                }, messages.displayError);
            }, messages.displayError);
        };
    }, messages.displayError);
});
