/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "aps/Tile",
    "aps/FieldSet",
    "./getResourceOutput",
    "./getStorageOutput",
    "./../widgets/base/Output",
    './../../common',
    "./../../messages",
    "./../utils",
    "dojo/_base/lang"
], function (
    Tile,
    FieldSet,
    getResourceOutput,
    getStorageOutput,
    Output,
    common,
    messages,
    utils,
    lang
) {
    var BACKUP_USER_STATUSES = {
        ERROR: "-1",
        INACTIVE: "0",
        ACTIVE: "1",
        IN_PROGRESS: "2",
        NOT_ACTIVATED: "3"
    };

    var BACKUP_SERVICE_STATUSES = {
        NOT_ASSIGNED: "not_assigned",
        PROVISIONING: 'provisioning',
        NOT_ACTIVATED: "not_activated",
        ACTIVATED: "activated",
        DISABLED: "disabled",
        UNPROVISIONING: 'unprovisioning',
        ERROR: "error"
    };

    function getBackupServiceStatus(backupAccount) {
        if (!backupAccount) {
            return BACKUP_SERVICE_STATUSES.NOT_ASSIGNED;
        }
        switch (backupAccount.aps.status) {
            case "aps:proto":
            case "aps:provisioning":
                return BACKUP_SERVICE_STATUSES.PROVISIONING;
            case "aps:unprovisioning":
            case "aps:deleted":
                return BACKUP_SERVICE_STATUSES.UNPROVISIONING;
            case "aps:ready":
                switch (backupAccount.userstatus) {
                    case BACKUP_USER_STATUSES.ERROR:
                        return BACKUP_SERVICE_STATUSES.ERROR;
                    case BACKUP_USER_STATUSES.INACTIVE:
                        return BACKUP_SERVICE_STATUSES.DISABLED;
                    case BACKUP_USER_STATUSES.ACTIVE:
                        return BACKUP_SERVICE_STATUSES.ACTIVATED;
                    case BACKUP_USER_STATUSES.NOT_ACTIVATED:
                        return BACKUP_SERVICE_STATUSES.NOT_ACTIVATED;
                }
        }
        return undefined;
    }

    function createUserResources(backupAccount, backupAccountStatistic, tenant, countersList) {
        var fs = new FieldSet({
            gridSize: "md-12 xs-12"
        });

        fs.addChild(new Output({
            gridSize: "md-12 xs-12",
            label: _("Login"),
            value: backupAccount.endUserLogin
        }));

        countersList.forEach(function (counter) {
            if (tenant[counter.name] && tenant[counter.name].limit && backupAccountStatistic &&
                backupAccountStatistic[counter.name] && backupAccountStatistic[counter.name].limit) {
                var tileData = lang.mixin({
                    gridSize: 'md-4 xs-12',
                    label: counter.label
                }, counter.legend);

                switch (counter.schema) {
                    case 'feature':
                        fs.addChild(new Output(lang.mixin({
                            value: _('Enabled')
                        }, tileData)));
                        break;
                    case 'storage':
                        fs.addChild(getStorageOutput(lang.mixin({
                            usage: backupAccountStatistic[counter.name].usage || 0,
                            limit: backupAccountStatistic[counter.name].limit || 0
                        }, tileData)));
                        break;
                    default:
                        fs.addChild(getResourceOutput(lang.mixin({
                            usage: backupAccountStatistic[counter.name].usage || 0,
                            limit: backupAccountStatistic[counter.name].limit || 0
                        }, tileData)));
                }
            }
        });

        return fs;
    }

    function createUserActivation() {
        return new Output({
            escapeHTML: false,
            gridSize: "md-12 xs-12",
            content: utils.ol([
                _("Click Activate button to set password for your service user."),
                _("Download and install the backup agent on your device. During the installation, specify your service user login and password."),
                _("Apply a backup plan to the device after the agent installation."),
                _("Log into Backup Console to manage backups.")
            ])
        });
    }

    function getUserResourcesTile(backupAccount, backupAccountStatistic, backupAccountActions, tenant, countersList) {
        var status = getBackupServiceStatus(backupAccount),
            buttons = [];

        var tile = new Tile({
            title: _("Service User"),
            gridSize: "md-8 xs-12"
        });

        switch (status) {
            case BACKUP_SERVICE_STATUSES.NOT_ASSIGNED:
                tile.addChild(new Output({
                    value: _("The service is not yet assigned. Click <b>Assign Service</b> button below to activate it."),
                    escapeHTML: false
                }));
                buttons.push({
                    iconClass: "fa fa-check",
                    title: _("Assign Service"),
                    autoBusy: false,
                    onClick: function () {
                        backupAccountActions.assign().then(null, messages.displayError);
                    }
                });
                break;

            case BACKUP_SERVICE_STATUSES.PROVISIONING:
                tile.addChild(new Output({value: _("Please wait...")}));
                break;

            case BACKUP_SERVICE_STATUSES.NOT_ACTIVATED:
                tile.addChild(createUserActivation());
                tile.set("title", _("Getting Started"));
                buttons.push({
                    iconClass: "fa fa-magic",
                    title: _("Activate"),
                    autoBusy: false,
                    onClick: function () {
                        backupAccountActions.activate().then(null, messages.displayError);
                    }
                });
                break;

            case BACKUP_SERVICE_STATUSES.ACTIVATED:
                tile.addChild(createUserResources(backupAccount, backupAccountStatistic, tenant, countersList));

                buttons.push({
                    iconClass: "fa fa-external-link",
                    title: _("Log into Backup Console"),
                    autoBusy: false,
                    onClick: function () {
                        var w = window.open('', '_blank');
                        backupAccountActions.loginToConsole().then(function (uri) {
                            w.location.href = uri;
                        }, function (error) {
                            w.close();
                            messages.displayError(error);
                        });
                    }
                });
                buttons.push({
                    iconClass: "fa fa-key",
                    title: _("Change Password"),
                    type: "danger",
                    autoBusy: false,
                    onClick: function () {
                        backupAccountActions.changePassword().then(null, messages.displayError);
                    }
                });
                break;

            case BACKUP_SERVICE_STATUSES.DISABLED:
                tile.addChild(new Output({value: _("Your account is disabled.")}));
                break;

            case BACKUP_SERVICE_STATUSES.UNPROVISIONING:
                tile.addChild(new Output({value: _("Please wait...")}));
                break;

            case BACKUP_SERVICE_STATUSES.ERROR:
                tile.addChild(new Output({value: _("Oops...")}));
                break;
        }

        tile.set("buttons", buttons);
        return tile;
    }

    return getUserResourcesTile;
});

