/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/promise/all",
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/Deferred",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "aps/_View",
    "aps/Memory",
    "aps/Container",
    "aps/Output",
    "aps/load",
    "./../components/getUserResourcesGrid",
    "./../../messages",
    "./../../common"
], function (
    all,
    declare,
    lang,
    Deferred,
    at,
    getStateful,
    getPlainValue,
    _View,
    Memory,
    Container,
    Output,
    load,
    getUserResourcesGrid,
    messages,
    common
) {
    return declare(_View, {
        _container: null,
        _tenant: null,
        _model: null,
        _userQuota: null,
        _deltas: [],
        _users: null,

        _submit: function () {
            var self = this,
                actions = {
                    provisionResource: [],
                    configureResource: []
                },
                users = self._users;

            if (Object.keys(self._model.notifications).length) {
                self._model.notifications = common.transformNotificationsToArray(self._model.notifications);
            } else {
                delete self._model.notifications;
            }
            var data = getPlainValue(self._model);

            users.forEach(function (user) {
                var backupAccount = lang.clone(data);
                if (user.backupUser) {
                    backupAccount.aps = {id: user.backupUser.aps.id};
                    actions.configureResource.push(backupAccount);
                } else {
                    backupAccount.user = {aps: {id: user.aps.id}};
                    backupAccount.aps = {
                        type: common.TYPES.USERS,
                        subscription: self._tenant.aps.subscription
                    };
                    backupAccount.tenant = [{aps: {id: self._tenant.aps.id}}];
                    actions.provisionResource.push(backupAccount);
                }
            });

            var operation = {
                "call": {
                    "resourceId": self._tenant.aps.id,
                    "operation": "setUsersLicensesAfterUpsell",
                    "parameters": {"data": actions}
                }
            };

            return aps.biz.requestUsageChange({
                deltas: self._deltas,
                operations: [operation],
                subscriptionId: self._tenant.aps.subscription
            }).then(function() {
                return aps.biz.commit();
            });
        },

        init: function () {
            this._container = new Container();
            this._container.placeAt(this.domNode);
        },
        onHide: function () {
            messages.removeAll();
        },
        onCancel: function () {
            aps.apsc.prev();
        },
        onSubmit: function () {
            var self = this;
            if (!self._container.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }

            self._submit().then(function () {
                aps.apsc.prev();
            }, messages.displayError);
        },
        onContext: function (context) {
            var self = this,
                usersIds = aps.context.params.usersIds,
                users = [],
                allUsers = context.vars.users;

            if (!usersIds) {
              return;
            }

            allUsers.forEach(function (user) {
                if (0 <= usersIds.indexOf(user.aps.id))
                    users.push(user);
            });

            self._container.removeAll();
            self.set("label", _("Assign Cyber Protection service for __count__ user(s)", {count: users.length}));
            self._container.addChild(new Output({
                value: _("To assign service you need to define resource limits. These limits will be applied for each user you previously selected.")
            }));

            self._tenant = context.vars.tenant;
            self._users = users;

            all({
                profile: common.getBackupProfile(self._tenant),
                offeringItems: common.getBackupAccountResources(self._tenant)
            }).then(function(response) {
                var profile = response.profile,
                    offeringItems = response.offeringItems;
                self._userQuota = common.getUserQuota(profile);
                self._model = getStateful({
                    notifications: common.transformNotificationsToObject(profile.notifications),

                    privileged: common.USER_PRIVILEGED.NO,

                    activateDuringCreation: false,

                    userUsage: {}
                });

                let offeringItemsNames = offeringItems.map(function(oi) { return window.offeringItemsMap[oi.property]; }),
                    limitsValue = self._userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE;

                self._model.set("offeringItems", offeringItemsNames.reduce(function (offeringItems, offeringItemName) {
                    offeringItems.push(getStateful({limit: limitsValue, name: offeringItemName}));
                    return offeringItems;
                }, []));

                var form = ['aps/Tiles', {title: _('Service User Settings')}, [
                    ['aps/Tile',{title: _('Management Portal')},  [
                        ['aps/CheckBox', {
                            description: _('Quota overuse notification'),
                            checked: at(self._model.notifications, 'quota')
                        }],
                        ['aps/CheckBox', {
                            description: _('Schedule usage report'),
                            checked: at(self._model.notifications, 'reports')
                        }]
                    ]],
                    ['aps/Tile', {title: _('Cyber Protection')}, [
                        ['aps/CheckBox', {
                            description: _('Failure notifications'),
                            checked: at(self._model.notifications, 'backup_error')
                        }],
                        ['aps/CheckBox', {
                            description: _('Warning notifications'),
                            checked: at(self._model.notifications, 'backup_warning')
                        }],
                        ['aps/CheckBox', {
                            description: _('Success notifications'),
                            checked: at(self._model.notifications, 'backup_info')
                        }],
                        ['aps/CheckBox', {
                            description: _('Daily recap about active alerts'),
                            checked: at(self._model.notifications, 'backup_daily_report')
                        }]
                    ]],
                    ['aps/Tile', {title: _('Privileges')}, [
                        ['aps/FieldSet', [
                            ['aps/CheckBox', {
                                description: _('Administrator'),
                                checked: at(self._model, 'privileged').transform({
                                    parse: function(value) {
                                        return value ? common.USER_PRIVILEGED.YES : common.USER_PRIVILEGED.NO;
                                    },
                                    format: function(value) {
                                        return value === common.USER_PRIVILEGED.YES;
                                    }
                                })
                            }],
                            ['aps/Output', {description: _('An administrator can manage backups of other users and back up Office 365 seats.')}]
                        ]]
                    ]]
                ]];

                load(form, self._container).then(function () {
                    if (self._userQuota) {
                        getUserResourcesGrid(self._tenant, self._model, users.length).then(function (resources) {
                            self._container.addChild(resources.grid);
                            self._deltas = resources.deltas;
                            aps.apsc.hideLoading();
                        }, messages.displayError);
                    }
                    else {
                        aps.apsc.hideLoading();
                    }
                }, messages.displayError);

            }, messages.displayError);
        }
    });
});



