/*
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require([
    "aps/load",
    "aps/Memory",
    "aps/ResourceStore",
    "aps/xhr",
    "dijit/registry",
    "dojo/promise/all",
    "dojox/mvc/at",
    "dojox/mvc/getPlainValue",
    "dojox/mvc/getStateful",
    "app/common",
    "app/messages",
    "app/storagesHelper"
], function (
    load,
    Memory,
    Store,
    xhr,
    registry,
    all,
    at,
    getPlainValue,
    getStateful,
    common,
    messages,
    storagesHelper
) {
    var globals = aps.context.vars.globals,
        profile = aps.context.vars.profile,
        profilesStore = new Store({
            target: "/aps/2/resources/"
        }),
        profileModel = getStateful({
            aps: {id: profile.aps.id},
            name: profile.name,
            datacenters: profile.datacenters || [],
            storages: profile.storages || [],
            storagePolicy: profile.storagePolicy || common.STORAGE_POLICY.PARENT,
            storageLocation: null,
            userQuota: profile.userQuota,
            freemiumMailboxStorage: profile.freemiumMailboxStorage,
            mspGroupType: profile.mspGroupType || common.GROUP_TYPE.CUSTOMER,
            notifications: common.transformNotificationsToObject(profile.notifications)
        }),
        isFreemiumMailboxStorageEnabled = globals.minimalCommitment; // Enabled only for Ingram

    all({
        datacenters: storagesHelper.loadDatacenters(globals),
        profileStatus: xhr.get("/aps/2/resources/" + profile.aps.id + "/get_status")
    }).then(function (response) {
        var datacentersSelectOptions = storagesHelper.makeDatacentersOptions(response.datacenters),
            profileDatacenters = storagesHelper.filterDatacenters(response.datacenters, profile),
            selectedDatacenters = getStateful(profileDatacenters.map(function (item) {
                return item.id;
            })),
            storageLocations = [
                {value: common.BACKUP_LOCATIONS.LOCAL_AND_CLOUD, label: _("Local and cloud backups")},
                {value: common.BACKUP_LOCATIONS.ONLY_CLOUD, label: _("Cloud only backups")}
            ],
            canRestrictQuota = profile.userQuota == common.USER_QUOTA.STRICT || !response.profileStatus.isUsed;
        storagesHelper.loadStorages(response.datacenters).then(function (storagesLoaded) {
            var storagesSelectOptions = storagesHelper.makeStoragesOptions(storagesLoaded),
                profileStorages = storagesHelper.filterStorages(storagesLoaded, profile),
                selectedStorages = getStateful(profileStorages.map(function (item) {
                    return item.datacenterId + '_' + item.storageId;
                }));

            selectedStorages.watchElements(function () {
                profileModel.set("storages", selectedStorages.map(function (value) {
                    var selectedOption = storagesSelectOptions.find(function (option) {
                        return option.value === value;
                    });
                    return {
                        datacenterId: selectedOption.datacenterId,
                        storageId: selectedOption.storageId
                    };
                }));
            });

            selectedDatacenters.watchElements(function () {
                profileModel.set("datacenters", selectedDatacenters.map(function (value) {
                    var selectedOption = datacentersSelectOptions.find(function (option) {
                        return option.value === value;
                    });
                    return {
                        datacenterId: selectedOption.value
                    };
                }));
            });

            var widgets = ["aps/PageContainer", {id: "apsPageContainer"}, [
                ["aps/FieldSet", {"id": common.genId('fieldset'), title: _("General")}, [
                    ["aps/TextBox", {
                        id: "name",
                        label: _("Profile name"),
                        value: at(profileModel, "name"),
                        trim: true,
                        required: true,
                        hint: _("The resource name will not be automatically changed if you change the profile name.")
                    }],
                    ["aps/Output", {
                        id: "resourceType",
                        label: _("Resource"),
                        value: profile.resourceTypeId
                    }]
                ]],
                ["aps/FieldSet", {id: common.genId('fieldset'), title: _("Options")}, [
                    ["aps/Select", {
                        id: "mspGroupType",
                        label: _("Account type"),
                        value: at(profileModel, "mspGroupType"),
                        required: true,
                        store: new Memory({
                            data: [
                                {value: common.GROUP_TYPE.CUSTOMER, label: _("Customer")},
                                {value: common.GROUP_TYPE.PARTNER, label: _("Service provider")}
                            ],
                            idProperty: "value"
                        })
                    }],
                    ["aps/Select", {
                        id: "storagePolicy",
                        label: _("Storage"),
                        value: at(profileModel, "storagePolicy"),
                        store: new Memory({
                            data: [
                                {value: common.STORAGE_POLICY.PARENT, label: _("Selected storage")},
                                {value: common.STORAGE_POLICY.OWN, label: _("Resellers must use their own storage")}
                            ], idProperty: "value"
                        }),
                        visible: at(profileModel, "mspGroupType").transform({
                            format: function (value) {
                                return value != common.GROUP_TYPE.PARTNER;
                            }
                        }).direction(at.from)
                    }],
                    ["aps/MultiSelect", {
                        id: "storages",
                        titleLeft: _("Available:"),
                        titleRight: _("Selected:"),
                        values: selectedStorages,
                        store: new Memory({data: storagesSelectOptions, idProperty: "value"})
                    }],
                    ["aps/MultiSelect", {
                        id: "datacenters",
                        label: _("Datacenter"),
                        titleLeft: _("Available:"),
                        titleRight: _("Selected:"),
                        values: selectedDatacenters,
                        store: new Memory({data: datacentersSelectOptions, idProperty: "value"})
                    }],
                    ["aps/Output", {
                        id: "storagesHint",
                        value: _("Changing this option will not impact the existing subscriptions."),
                        'class': "hint",
                        visible: at(profileModel, "storagePolicy").transform({
                            format: function (value) {
                                return value != common.STORAGE_POLICY.OWN;
                            }
                        }).direction(at.from)
                    }],
                    ["aps/Select", {
                        id: "storageLocation",
                        label: _("Backup locations"),
                        value: profile.storageLocation,
                        disabled: true,
                        store: new Memory({data: storageLocations, idProperty: "value"}),
                        visible: profile.storageLocation != null
                    }],
                    ["aps/CheckBox", {
                        id: "quotaEnabled",
                        label: _("Manage user quotas"),
                        checked: at(profileModel, "userQuota").transform({
                            format: function (value) {
                                return value != common.USER_QUOTA.OFF;
                            },
                            parse: function (value) {
                                return value ? (canRestrictQuota ? common.USER_QUOTA.STRICT : common.USER_QUOTA.SOFT) :
                                    common.USER_QUOTA.OFF;
                            }
                        }),
                        visible: at(profileModel, "mspGroupType").transform({
                            format: function (value) {
                                return value != common.GROUP_TYPE.PARTNER;
                            }
                        }).direction(at.from)
                    }],
                    ["aps/CheckBox", {
                        id: "overLimit",
                        label: _("Do not allow to set user quotas exceeding subscription limits"),
                        hint: canRestrictQuota ? "" : _("You cannot set this option because some of the user quotas already exceed the subscription limits."),
                        checked: at(profileModel, "userQuota").transform({
                            format: function (value) {
                                return value == common.USER_QUOTA.STRICT;
                            },
                            parse: function (value) {
                                return value ? common.USER_QUOTA.STRICT :
                                    (registry.byId("quotaEnabled").checked ? common.USER_QUOTA.SOFT : common.USER_QUOTA.OFF);
                            }
                        }),
                        disabled: at(profileModel, "userQuota").transform({
                            format: function (value) {
                                return value == common.USER_QUOTA.OFF || !canRestrictQuota;
                            }
                        }).direction(at.from),
                        visible: at(profileModel, "mspGroupType").transform({
                            format: function (value) {
                                return value != common.GROUP_TYPE.PARTNER;
                            }
                        }).direction(at.from)
                    }],
                    ["aps/TextBox", {
                        id: "freemiumMailboxStorage",
                        label: _("Free cloud storage per Microsoft Office 365 seat (GB)"),
                        trim: true,
                        placeHolder: _('Unlimited'),
                        pattern: "[0-9]*",
                        value: at(profileModel, "freemiumMailboxStorage").transform({
                            format: function (value) {
                                return value == common.QUOTA_UNLIMITED_VALUE ?
                                    '' :
                                    Math.round(Number(value) / common.BYTES_IN_GB);
                            },
                            parse: function (value) {
                                return value === '' ?
                                    common.QUOTA_UNLIMITED_VALUE :
                                    Number(value) * common.BYTES_IN_GB;
                            }
                        }),
                        visible: at(profileModel, "mspGroupType").transform({
                            format: function (value) {
                                return value != common.GROUP_TYPE.PARTNER && isFreemiumMailboxStorageEnabled;
                            }
                        }).direction(at.from)
                    }]
                ]],
                ["aps/FieldSet", {"id": common.genId('fieldset'), title: _("Notifications")}, [
                    ["aps/FieldSet", {
                        label: "MANAGEMENT PORTAL",
                        title: false
                    }],
                    ["aps/CheckBox", {
                        label: _('Quota overuse notifications'),
                        checked: at(profileModel.notifications, "quota")
                    }],
                    ["aps/CheckBox", {
                        label: _('Scheduled usage reports'),
                        checked: at(profileModel.notifications, "reports")
                    }],
                    ["aps/Output", {
                        class: 'space-before-element'
                    }],
                    ["aps/FieldSet", {
                        label: "CYBER PROTECTION",
                        title: false
                    }],
                    ["aps/CheckBox", {
                        label: _('Failure notifications'),
                        checked: at(profileModel.notifications, "backup_error")
                    }],
                    ["aps/CheckBox", {
                        label: _('Warning notifications'),
                        checked: at(profileModel.notifications, "backup_warning")
                    }],
                    ["aps/CheckBox", {
                        label: _('Success notifications'),
                        checked: at(profileModel.notifications, "backup_info")
                    }],
                    ["aps/CheckBox", {
                        label: _('Daily recap about active alerts'),
                        checked: at(profileModel.notifications, "backup_daily_report")
                    }]
                ]]
            ]];

            load(widgets).then(function () {
                if (profile.storageLocation != null) {
                    messages.displayWarning(
                        _('Property "{backup_locations}" is deprecated and will be removed after clicking "{submit}".')
                            .replace('{backup_locations}', _('Backup locations'))
                            .replace('{submit}', _('Submit'))
                    );
                }
                function refreshVisible() {
                    var storagesSelect = registry.byId("storages"),
                        datacentersSelect = registry.byId("datacenters"),
                        showDatacenters = profileModel.mspGroupType == common.GROUP_TYPE.PARTNER,
                        showStorages = !showDatacenters && profileModel.storagePolicy != common.STORAGE_POLICY.OWN;

                    datacentersSelect.set("visible", showDatacenters);
                    storagesSelect.set("visible", showStorages);
                }
                refreshVisible();
                profileModel.watch(refreshVisible);

                aps.app.onSubmit = function () {
                    var page = registry.byId("apsPageContainer");
                    if (!page.validate()) {
                        aps.apsc.cancelProcessing();
                        return;
                    }
                    if (Object.keys(profileModel.notifications).length) {
                        profileModel.notifications = common.transformNotificationsToArray(profileModel.notifications, []);
                    } else {
                        delete profileModel.notifications;
                    }
                    var profileData = getPlainValue(profileModel);
                    if (profileModel.mspGroupType == common.GROUP_TYPE.PARTNER) {
                        if (profileModel.datacenters.length === 0) {
                            messages.displayWarning(_("You have to choose a datacenter."));
                            aps.apsc.cancelProcessing();
                            return;
                        }
                        profileData.storages = [];
                    } else {
                        if (profileModel.storagePolicy == common.STORAGE_POLICY.PARENT && profileModel.storages.length === 0) {
                            messages.displayWarning(_("You have to choose a storage."));
                            aps.apsc.cancelProcessing();
                            return;
                        }
                        if (profileData.storagePolicy == common.STORAGE_POLICY.OWN) {
                            profileData.storages = [];
                        }
                        profileData.datacenters = [];
                    }

                    profilesStore.put(profileData).then(function () {
                        aps.apsc.gotoView(common.getRelatedViewId('view'), profile.aps.id);
                    }, messages.displayError);
                };
            }, messages.displayError);
        });
    }, messages.displayError);

    aps.app.onCancel = function () {
        aps.apsc.gotoView(common.getRelatedViewId('view'), profile.aps.id);
    };
});
