/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

/**
 * Renders view user page controls.
 */
define([
    "dojo/Deferred",
    "aps/i18n",
    "aps/xhr",
    "app/common",
    "app/quotas"
], function (
    Deferred,
    _,
    xhr,
    common,
    quotas
) {
    function getOverviewValue(counterField) {
        return counterField == common.QUOTA_UNLIMITED_VALUE ? _('Unlimited') : counterField;
    }

    function getStorageOverview(storageLimit) {
        return storageLimit == common.QUOTA_UNLIMITED_VALUE ? _('Unlimited') : storageLimit / common.BYTES_IN_GB;
    }

    function printUserStatus(status) {
        switch (status) {
            case common.USER_STATUS.ACTIVE:
                return _('Enabled');
            case common.USER_STATUS.IN_PROGRESS:
                return _('In progress');
            case common.USER_STATUS.INACTIVE:
                return _('Disabled');
            case common.USER_STATUS.NOT_ACTIVATED:
                return _('Not activated');
            default:
                return _('Disabled');
        }
    }

    function printNotification(notifications, value) {
        return notifications.includes(value) ? _('Yes') : _('No');
    }

    function renderControls(account, user, showBack, userQuota, tenant, statistics) {
        var showQuotas = quotas.canShowQuotas(userQuota, account),
            countersList = common.getAvailableCountersForArea('quota_controls_ccpv1_view_user'),
            quotaOutputs = [];

        if (showQuotas) {
            for (let counter of countersList) {
                if (!(tenant[counter.name] && tenant[counter.name].limit)) {
                    continue;
                }

                switch (counter.schema) {
                    case 'feature':
                        quotaOutputs.push(["aps/Output", {
                            label: counter.label,
                            value: _('Enabled')
                        }]);
                        break;
                    case 'storage':
                        if(statistics.hasOwnProperty(counter.name)) {
                            counter.label = quotas.formatStorageLabel(counter.label, statistics[counter.name].limit || 0)
                            quotaOutputs.push(["aps/Output", {
                                label: counter.label,
                                value: getStorageOverview(statistics[counter.name].limit || 0)
                            }]);
                        }
                        break;
                    default:
                        if(statistics.hasOwnProperty(counter.name)) {
                            quotaOutputs.push(["aps/Output", {
                                label: counter.label,
                                value: getOverviewValue(statistics[counter.name].limit || 0)
                            }]);
                        }
                }
            }
        }

        return ["aps/PageContainer", {id: "apsPageContainer"}, [
            ["aps/FieldSet", {
                title: _("Main")
            }, [
                ["aps/Output", {
                    label: _("User name"),
                    value: common.escapeAngularBrackets(account.displayName + ' (' + account.endUserLogin + ')')
                }],
                ["aps/Output", {
                    label: _('Administrator privileges'),
                    value: account.privileged ? _('Yes') : _('No')
                }]
            ]],
            ["aps/FieldSet", {
                escapeHTML: false,
                title: _('Account status')
            }, [
                ["aps/Output", {
                    label: _('Status'),
                    value: printUserStatus(account.userstatus)
                }]
            ]],
            ["aps/FieldSet", {
                escapeHTML: false,
                title: _('Notifications')
            }, [
                ["aps/FieldSet", {
                    label: "MANAGEMENT PORTAL",
                    title: false
                }],
                ["aps/Output", {
                    label: _('Quota overuse notifications'),
                    value: printNotification(account.notifications, 'quota')
                }],
                ["aps/Output", {
                    label: _('Scheduled usage reports'),
                    value: printNotification(account.notifications, 'reports')
                }],
                ["aps/Output", {
                    class: 'space-before-element'
                }],
                ["aps/FieldSet", {
                    label: "CYBER PROTECTION",
                    title: false
                }],
                ["aps/Output", {
                    label: _('Failure notifications'),
                    value: printNotification(account.notifications, 'backup_error')
                }],
                ["aps/Output", {
                    label: _('Warning notifications'),
                    value: printNotification(account.notifications, 'backup_warning')
                }],
                ["aps/Output", {
                    label: _('Success notifications'),
                    value: printNotification(account.notifications, 'backup_info')
                }],
                ["aps/Output", {
                    label: _('Daily recap about active alerts'),
                    value: printNotification(account.notifications, 'backup_daily_report')
                }]
            ]],
            ["aps/FieldSet", {
                escapeHTML: false,
                title: _('Quotas'),
                visible: showQuotas
            }, quotaOutputs],
            ["aps/FieldSet", {title: true}],
            ["aps/Container", {id: "buttons"}, [
                ["aps/Button", {
                    title: _('Edit'),
                    disabled: user.disabled,
                    onClick: function () {
                        var viewId = common.getCurrentView() === 'ccp--suservice--view' ?
                            'ccp--user--edit' :
                            common.getRelatedViewId('edit');
                        aps.apsc.gotoView(viewId, account.aps.id);
                    }
                }],
                ["aps/Button", {
                    title: _('Back to user list'),
                    visible: showBack,
                    onClick: function () {
                        aps.apsc.gotoView(common.getRelatedViewId('list'));
                    }
                }],
                ["aps/Button", {
                    title: _('Change Password'),
                    visible: account.userstatus == common.USER_STATUS.ACTIVE && !account.activateDuringCreation,
                    onClick: function () {
                        aps.apsc.gotoView(common.getRelatedViewId('change-password'), account.aps.id);
                    }
                }]
            ]]
        ]];
    }

    return {
        renderControls: renderControls
    };
});

