<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\ApplicationCounter;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Services\Redis\RedisAwareTrait;

/**
 * Class Model_Resource_Tenant_Customer
 */
class Model_Resource_Tenant_Customer extends Model_Resource_Tenant
{
    use RedisAwareTrait;
    use CountersManagerAwareTrait;

    const LOCK_USER_LIMITS_CALCULATION_TTL = 300;

    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->checkProfile();
        $this->linkProfile(false);

        $tenant = $this->tenant;
        $tenant->datacenterId = $this->resolveDatacenterId();
        $this->linkDatacenter(false);

        $this->updateCounterLimits();

        if ($tenant->ismigration && !is_null($tenant->mspGroupId)) {
            Log::getInstance()->event(self::LOG_MIGRATION_WAS_STARTED);
            $this->_migrateMspGroup();

            $this->_retrieveUsageForCustomer();
            $this->_updateGroupForCustomer();
        } else {
            $tenant->mspGroupId = $this->_createGroupForCustomer();
        }
        $this->updateTenantId();
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateStartProductionTime();

        $this->linkDatacenter(false);
        $this->linkProfile(false);

        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->updateCounterLimits();
        $this->_retrieveUsageForCustomer();
    }

    public function configure($new)
    {
        $this->lockUserLimitsCalculation();

        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;

        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }

        $this->_updateGroupForCustomer();

        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        $this->obtainUsersLimitsFromCloud();
        $this->releaseUserLimitsCalculation();
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            Log::getInstance()->event(self::LOG_UNPROVISIONING_TENANT_IS_MIGRATION);
            return;
        }

        if (is_null($tenant->mspGroupId)) {
            Log::getInstance()->warning(self::LOG_WARNING_UNPROVISION_TENANT_DOES_NOT_HAVE_GROUP);
            return;
        }
        Log::getInstance()->event(sprintf(self::LOG_FORMAT_DELETE_MSP_GROUP, $tenant->account->companyName));

        $this->linkDatacenter(false);

        try {
            $this->_deleteGroupForCustomer();
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                throw $e;
            }
            sleep(10);
            $this->_deleteGroupForCustomer();
        }
    }

    public function enable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();
        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->getTenantManager()->enableGroup($tenant->mspGroupId);
    }

    public function disable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->linkProfile(false);
        $this->linkDatacenter(false);
        $this->getTenantManager()->disableGroup($tenant->mspGroupId);
    }

    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->datacenterId)) {
            $tenant->datacenterId = globals::DEFAULT_DATACENTER_ID;
        }

        $this->updateStartProductionTime();
        $this->updateTenantId();
        $this->update();

        $realLink = !$tenant->subscription->disabled;
        if ($realLink) {
            $this->checkProfile();
        }

        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);
        $this->_getOrCreateParentGroupId($tenant->datacenter);  // add parents in map
        $this->_updateGroupForCustomer();
        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this, true);
        $this->_upgradeUsers();

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    /* ======= RESOURCE EVENTS ======= */

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();

        if ($this->tenant->ismigration) {
            return;
        }

        $applicationConfig = $this->getApplicationConfig();
        if (!$applicationConfig->service_auto_assignation) {
            return;
        }
        $admin = $this->getApsApi()->getAdmin($this->tenant);
        if (is_null($admin)) {
            return;
        }
        $this->provisionDefaultUser($admin);
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $needUpdate = $this->updateStartProductionTime();
        if ($this->updateCounterLimits()) {
            $needUpdate = true;
        }
        if ($needUpdate) {
            $this->update();
        }
        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);
        $this->_updateGroupForCustomer();
        if (!$tenant->subscription->disabled) {
            $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);
        }

        if ($this->canCalculateUserLimits()) {
            $userQuota = $this->_getProfileUserQuota();
            $tenant = $this->tenant;
            $availableOfferingItemsHashmap = $this->getCloudApi()->getOfferingItemsAvailableForChildHashMap($tenant->tenantId);
            foreach ($tenant->users as $user) {
                $this->updateUserApsResource($tenant, $user, $userQuota, $availableOfferingItemsHashmap);

                try {
                    $this->updateUserCloudAccount($user);
                } catch (Service_BaseClient_HTTPException $e) {
                    if ($e->getCode() !== Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                        throw $e;
                    }
                    $this->updateUserCloudAccount($user);
                }
            }
        }
    }

    public function onUserChanged($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    public function onAccountChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @return array
     * @throws Exception
     */
    public function getStorages()
    {
        Log::getInstance()->event(
            self::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        return $this->memoize(function () {
            $tenant = $this->tenant;
            $response = [];
            $storages = $this->getMspApi()->getStorages($tenant->mspGroupId);
            if (is_null($storages)) {
                Log::getInstance()->error(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
                throw new Exception(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
            }

            foreach ($storages as $storage) {
                /** @var Service_MspApiClient_Response_Struct_Storage $storage */
                if ($storage->getInherited()) {
                    continue;
                }
                $response[] = [
                    self::FIELD_STORAGE_ID => (string)$storage->getId(),
                    self::FIELD_STORAGE_NAME => $storage->getName(),
                ];
            }

            return $response;
        });
    }

    /**
     * @return object
     */
    public function getDatacenter()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkDatacenter($realLink);
        return (object)$tenant->datacenter->toArray();
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     */
    public function loginExists($login, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        // Is the user's parent the same tenant?
        $user = $this->findUserByLogin($this->tenant->mspGroupId, $login);
        if (empty($user)) {
            // Does the user exist in Cloud
            return $this->checkLoginExist($login, $datacenterId);
        }

        // Is the user already related to an OA user?
        $rql = RqlBuilder::createInstance()
            ->implementing(APSResource::CLOUD_BACKUP_USERS_TYPE)
            ->eq('endUserLogin', $login)
            ->limit(0, 1)
            ->getQuery();
        $resources = $this->getApsc()->impersonate($this->tenant->global)->getResources($rql);
        return !empty($resources);
    }

    /**
     * @return array
     */
    public function getQuotasStatistics()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $quotasStatistics = array();
        $tenant = $this->tenant;
        foreach ($this->fetchSubscriptionCounterNames() as $counterName) {
            $allocatedTotal = 0;
            $rql = RqlBuilder::createInstance()
                ->implementing(APSResource::CLOUD_BACKUP_USERS_TYPE)
                ->eq('tenant.aps.id', $tenant->aps->id)
                ->getQuery();
            $users = $this->getApsc()->getResources($rql);
            foreach ($users as $user) {
                /** @var users $user */
                /** @var UserOfferingItem|null $offeringItem */
                $offeringItem = $this->getUserOfferingItemByCounterName($user, $counterName);

                if (is_null($offeringItem) || !isset($offeringItem->limit)) {
                    continue;
                }
                if ($offeringItem->limit == self::UNLIMITED_VALUE) {
                    $allocatedTotal = self::UNLIMITED_VALUE;
                    break;
                }
                $allocatedTotal += $offeringItem->limit;
            }

            $limit = isset($tenant->{$counterName}->limit) ? $tenant->{$counterName}->limit : 0;

            $statistics = compact(
                'allocatedTotal',
                'limit'
            );

            // For backward compatibility only
            $quotasStatistics[lcfirst($counterName)] = $statistics;
            $quotasStatistics[$counterName] = $statistics;
        }
        return $quotasStatistics;
    }

    public function updateUsageAndLimits()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->updateCounterLimits();
        $this->_updateProfileLink();

        $tenantUsages = $this->fetchTenantLimitsAndUsages();

        // ToDo: Implement to retrieve statistics about user via baseUsers::retrieve & baseUsers::update
        $tenant = $this->tenant;
        $usersUsages = $this->getUserUsages($tenant);

        return [
            static::TENANT => $tenantUsages,
            static::USER_USAGE => $usersUsages,
            static::ERROR => null,
        ];
    }

    public function createAdministrator($parameters, $tenantId = '', $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrators()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function updateAdministrator($adminId, $parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function deleteAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getUserGuideUrl()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministratorConsole($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function resendAdministratorLink($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function enableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function disableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function adminUrlRedirect()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getDatacenters()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlans()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function setStorageForPlan($planId, $parameters)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function unsetStorageForPlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    /* ======= OTHER METHODS ======= */

    protected function createDefaultUserResource($tenant, $serviceUser)
    {
        $user = \APS\TypeLibrary::newResourceByTypeId(\APS\TypeLibrary::getTypeByServiceId('users'));
        $user->user->set($serviceUser);
        $user->tenant->set($tenant);
        $user->privileged = users::ADMINISTRATOR_PRIVILEGES_YES;
        $user->activateDuringCreation = false;
        $this->initUserFormProfile($user);

        return $user;
    }

    /**
     * @param $serviceUser
     */
    protected function provisionDefaultUser($serviceUser)
    {
        $tenant = $this->tenant;

        $user = $this->createDefaultUserResource($tenant, $serviceUser);

        $apsc = $this->getApsc()->impersonate($tenant);
        $apsc->resetSession();
        $apsc->provisionResource($user);
    }

    /**
     * @return object
     */
    protected function getApplicationConfig()
    {
        return Settings::getInstance()->getApplicationConfig();
    }

    public function resolveDatacenterId()
    {
        $tenant = $this->tenant;
        // Do nothing if datacenter ID already was set from external side (online store / migration script)
        if (isset($tenant->datacenterId)) {
            return $tenant->datacenterId;
        }

        $datacenters = $tenant->global->datacenters;
        if ($datacenters->count() === 0) {
            throw new \Rest\RestException(401, static::EXCEPTION_NEED_TO_SET_CREDENTIALS_MESSAGE);
        }

        $profile = $tenant->profile;
        $storage = $profile->storagePolicy == profile::STORAGE_POLICY_PARENT
            ? $this->resolveProviderStorage($profile)
            : $this->resolveResellerStorage();

        return $storage->datacenterId;
    }

    public function resolveBackupStorageId()
    {
        $tenant = $this->tenant;

        $profile = $tenant->profile;
        $storage = $profile->storagePolicy == profile::STORAGE_POLICY_PARENT
            ? $this->resolveProviderStorage($profile)
            : $this->resolveResellerStorage();

        if ($tenant->datacenterId != $storage->datacenterId) {
            throw new \Rest\RestException(409, Str::format(
                'Please select storage from datacenter "%s" in the Service Profile "%s".',
                $tenant->datacenterId, $profile->resourceTypeId
            ));
        };

        return $storage->storageId;
    }

    /**
     * Todo finish refactoring
     * @param $tenant
     * @return array
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function getUserUsages($tenant)
    {
        $userQuota = $this->_getProfileUserQuota();
        $mspUsers = $this->getMspApi()->getUsers($tenant->mspGroupId);
        $availableOfferingItemsToCountersMapping = Arr::map(
            $this->fetchSubscriptionCounters(),
            function($counter) {
                /** @var ApplicationCounter $counter */
                return $counter->getOfferingItemName();
            },
            function($counter) {
                /** @var ApplicationCounter $counter */
                return $counter->getName();
            },
            false
        );
        $tenant = $this->tenant;
        $cloudApi = $this->getCloudApi();
        $users = $cloudApi->getTenantUsers($tenant->tenantId);
        $usages = [];
        foreach ($users as $userId) {
            $user = $cloudApi->getUser($userId);
            $userLogin = $user->getLogin();
            $usages[$userLogin] = [
                static::USAGE => [],
                static::LIMIT => []
            ];
            $userUsages = $cloudApi->getTenantUsages(array_first($user->getPersonalTenantId()));
            foreach ($userUsages as $usage) {
                $usageName = $usage->getName();
                if (!isset($availableOfferingItemsToCountersMapping[$usageName])) {
                    continue;
                }
                $offeringItem = $usage->getOfferingItem();
                $offeringItemQuota = $offeringItem->hasQuota() && $offeringItem->getQuota()->hasValue()
                    ? $offeringItem->getQuota()->getValue()
                    : self::UNLIMITED_VALUE;
                $usages[$userLogin][static::USAGE][$usageName] =
                    $usage->hasValue() ? $usage->getValue() : 0;
                $usages[$userLogin][static::LIMIT][$usageName] =
                    $offeringItem->getStatus() === ApiInterface::OFFERING_ITEM_STATUS_ACTIVE ? $offeringItemQuota : 0;
            }
        }

        $usersUsages = [];
        if (is_array($mspUsers)) {
            foreach ($mspUsers as $userInfo) {
                $item = [];
                $login = $userInfo->getLogin();

                if (!isset($usages[$login])) {
                    $usersUsages[$userInfo->getId()] = [];
                    continue;
                }

                $userUsage = $usages[$login][static::USAGE];
                $userLimits = $usages[$login][static::LIMIT];
                $tenantArray = $tenant->toArray();
                foreach ($availableOfferingItemsToCountersMapping as $offeringItemName => $counterName) {
                    $item[$counterName] = [
                        static::USAGE =>Arr::get($userUsage, $offeringItemName),
                        static::LIMIT => $this->_getCounterLimit(
                            $userQuota,
                            Arr::getNestedItem($tenantArray, [$counterName, 'limit'], null),
                            Arr::get($userLimits, $offeringItemName)
                        ),
                    ];
                }

                $usersUsages[$userInfo->getId()] = $item;
            }
        }

        return $usersUsages;
    }

    protected function lockUserLimitsCalculation()
    {
        $this->getRedis()->set($this->getUserLimitsRedisKey(), true, static::LOCK_USER_LIMITS_CALCULATION_TTL);
    }

    protected function releaseUserLimitsCalculation()
    {
        $this->getRedis()->deleteAll($this->getUserLimitsRedisKey());
    }

    protected function canCalculateUserLimits()
    {
        return $this->getRedis()->has($this->getUserLimitsRedisKey());
    }

    private function getUserLimitsRedisKey()
    {
        return 'lock_user_limits_' . $this->tenant->tenantId;
    }

    /**
     * @param $profile
     * @return object
     * @throws \Rest\RestException
     */
    private function resolveProviderStorage($profile)
    {
        $tenant = $this->tenant;
        $storages = is_array($profile->storages) ? $profile->storages : [];
        $storagesCount = count($storages);
        if ($storagesCount === 0) {
            throw new \Rest\RestException(400, self::EXCEPTION_SELECT_STORAGE_IN_BACKUP_PROFILE_MESSAGE);
        }

        if ($storagesCount === 1) {
            return reset($storages);
        }

        $subscriptionParametersManager = $this->getSubscriptionParametersManager();
        $storageOptionId = $subscriptionParametersManager->getParameterStorageId($tenant->subscription->subscriptionId);

        foreach ($storages as $storage) {
            if ($storageOptionId != datacenter::calculateStorageOptionId($storage->datacenterId, $storage->storageId)) {
                continue;
            }

            return $storage;
        }

        throw new \Rest\RestException(400, self::EXCEPTION_UNKNOWN_STORAGE_MESSAGE);
    }

    private function resolveResellerStorage()
    {
        $tenant = $this->tenant;
        $parentTenant = $this->_getParentTenant($tenant);
        if (is_null($parentTenant)) {
            return null;
        }
        $planId = $this->_getServicePlanId();
        $storage = $this->_getStorageForPlan($parentTenant, $planId);

        if (is_null($storage)) {
            throw new Exception(self::EXCEPTION_RESELLER_HAS_TO_SET_STORAGE_FOR_PLAN_MESSAGE);
        }

        return (object)[
            'datacenterId' => $storage[self::FIELD_STORAGE_DATACENTER_ID],
            'storageId' => $storage[self::FIELD_STORAGE_ID],
        ];
    }

    /**
     * @param $tenant
     * @return \APS\ResourceBase
     */
    private function _getParentTenant($tenant)
    {
        $accountId = $this->_getParentAccountId($tenant->accountId);
        $tenants = $this->getApsApi()->getTenantsByAccountId($accountId);
        return count($tenants) ? $tenants[0] : null;
    }

    /**
     * @param int $accountId
     * @return int
     */
    private function _getParentAccountId($accountId)
    {
        $openApi = $this->getOpenApi();
        $accountInfo = $openApi->getAccountInfo($accountId);
        return $accountInfo->getParentAccountId();
    }

    /**
     * @return int
     * @throws Exception | Service_PbaApi_Api_Exception
     */
    private function _getServicePlanId()
    {
        $tenant = $this->tenant;
        $pbaApi = $this->getPbaApi();
        $subscriptionDetails = $pbaApi->getSubscriptionDetails($tenant->subscription->subscriptionId);
        return $subscriptionDetails->getPlanId();
    }

    /**
     * @throws \Acronis\App\UsageReport\Exception\ManagerException
     * @throws \Rest\RestException
     */
    private function _retrieveUsageForCustomer()
    {
        $this->updateCountersUsage($this->tenant->datacenter);
    }

    /**
     * @return void
     */
    private function _updateGroupForCustomer()
    {
        $tenant = $this->tenant;
        $mspGroupManager = $this->getMspGroupManager();
        $mspGroupManager->updateGroupForCustomer($tenant->mspGroupId, $tenant);
        $this->_lockGroup($tenant->mspGroupId);
    }

    private function _lockGroup($groupId)
    {
        try {
            $this->getMspApi()->lockGroup($groupId);
        } catch (Exception $e) {
            Log::getInstance()->error(sprintf(self::ERROR_FAILED_LOCK_GROUP, $groupId));
        }
    }

    /**
     * @return int
     */
    private function _createGroupForCustomer()
    {
        $tenant = $this->tenant;
        $parentMspGroupId = $this->_getOrCreateParentGroupId();
        $mspGroupManager = $this->getMspGroupManager();
        $groupId = $mspGroupManager->createGroupForCustomer($tenant, $parentMspGroupId);
        $this->_lockGroup($groupId);
        return $groupId;
    }

    private function obtainUsersLimitsFromCloud()
    {
        $tenant = $this->tenant;
        $applicationId = $this->getCloudApi()->getApplicationByType(ApiInterface::APPLICATION_TYPE_BACKUP)->getId();
        $userQuota = $this->_getProfileUserQuota();
        $allCounters = $this->getCountersManager()->getAllCountersMap();

        foreach ($tenant->users as $user) {
            $this->obtainUserLimitsFromCloud($user, $applicationId, $userQuota, $allCounters);

            try {
                $this->updateUserCloudAccount($user);
            } catch (Service_BaseClient_HTTPException $e) {
                if ($e->getCode() !== Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                    throw $e;
                }
                $this->updateUserCloudAccount($user);
            }
        }
    }

    /**
     * @param tenant $tenant
     * @param users $user
     * @param $userQuota
     * @param array $availableOfferingItemsHashmap
     * @throws Exception
     */
    private function updateUserApsResource(tenant $tenant, $user, $userQuota, $availableOfferingItemsHashmap)
    {
        $update = false;
        $counters = $this->getCountersManager()->getAllCounters();
        $offeringItems = $user->offeringItems;

        foreach ($counters as $counter) {
            if ($counter->getType() !== ApplicationCounter::TYPE_COUNTER) {
                continue;
            }
            $tenantCounterName = $counter->getName();
            $offeringItemName = $counter->getOfferingItemName();
            $offeringItem = Arr::find($offeringItems, function($oi) use ($offeringItemName) {
                return $oi->name === $offeringItemName;
            });
            $hasTenantLimit = !empty($tenant->{$tenantCounterName}->limit);

            if (is_null($offeringItem) && $hasTenantLimit) {
                $newOfferingItem = new UserOfferingItem();
                $newOfferingItem->name = $offeringItemName;
                $availableOfferingItem = Arr::get($availableOfferingItemsHashmap, $offeringItemName);
                if ($userQuota && $availableOfferingItem
                    && $availableOfferingItem->hasQuota()
                    && $availableOfferingItem->getQuota()->hasValue()) {
                    $newOfferingItem->limit = $availableOfferingItem->getQuota()->getValue();
                } elseif ($availableOfferingItem && $availableOfferingItem->getStatus() !== 0) {
                    $newOfferingItem->limit = self::UNLIMITED_VALUE;
                } else {
                    $newOfferingItem->limit = $userQuota ? 0 : self::UNLIMITED_VALUE;
                }

                $offeringItems[] = $newOfferingItem;
                $update = true;
            }

            if ($offeringItem) {
                if (!$hasTenantLimit && $offeringItem->limit != 0) {
                    $offeringItem->limit = 0;
                    $update = true;
                } elseif (!$userQuota && $offeringItem->limit != self::UNLIMITED_VALUE) {
                    $offeringItem->limit = self::UNLIMITED_VALUE;
                    $update = true;
                }
            }
        }
        if ($update) {
            /** @var \APS\ControllerProxy $apsc */
            $apsc = $tenant->getAPSC();
            $user->offeringItems = $offeringItems;
            $apsc->updateResource($user->resource);
        }
    }

    /**
     * @param $user
     * @throws Service_BaseClient_HTTPException
     */
    private function updateUserCloudAccount($user)
    {
        try {
            $userManager = $this->createUserManagerInstance($this->getMspApi());
            $userManager->updateUserPersonalQuotas($user);
        } catch (Service_BaseClient_HTTPException $e) {
            $skipCodes = [
                Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_FOUND,
                Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_AVAILABLE,
            ];
            if (!in_array($e->getCode(), $skipCodes)) {
                throw $e;
            }
            Log::getInstance()->warning(
                'Unable to update user "%s". Error: "%s".',
                $user->mspUserId, $e->getMessage()
            );
        }
    }

    /**
     * @return bool
     */
    private function _getProfileUserQuota()
    {
        $tenant = $this->tenant;
        $profile = $tenant->profile;
        if ($profile->isEmpty()) {
            $userQuota = true;
        } else {
            $userQuota = $profile->userQuota !== Service_APSApi_Api::PROFILE_USER_QUOTA_OFF;
        }
        return $userQuota;
    }


    /**
     * @throws Service_BaseClient_HTTPException
     */
    private function _deleteGroupForCustomer()
    {
        $mspApi = $this->getMspApi();
        $tenant = $this->tenant;
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        if (is_null($groupInfo)) {
            Log::getInstance()->warning(sprintf(self::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $tenant->mspGroupId));
            return;
        }
        $version = $groupInfo->getVersion();
        if ($groupInfo->getStatus() == 1) {
            $version = $this->getTenantManager()->disableGroup($tenant->mspGroupId, $version);
        }
        $mspApi->deleteGroup($tenant->mspGroupId, $version);
    }

    private function _getCounterLimit($userQuota, $tenantCounterLimit, $userCounterLimit)
    {
        if ($userQuota || $tenantCounterLimit) {
            return $userCounterLimit;
        }
        return $tenantCounterLimit;
    }

    private function _upgradeUsers()
    {
        $tenant = $this->tenant;
        $mspApi = $this->getMspApi();
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        $status = $groupInfo->getStatus();
        $version = $groupInfo->getVersion();
        $mspGroupManager = $this->getTenantManager();
        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $version = $mspGroupManager->enableGroup($tenant->mspGroupId, $version);
        }

        $exception = null;
        try {
            $users = $tenant->users;
            if (!count($users)) {
                $users = $this->getApsApi()->getUsersForTenant($tenant);
            }
            foreach ($users as $resource) {
                $this->upgradeUser($resource->aps->id);
            }
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            $exception = $e;
        }

        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $mspGroupManager->disableGroup($tenant->mspGroupId, $version);
        }

        if (!is_null($exception)) {
            throw $exception;
        }
    }

    /**
     * @param $userId
     */
    private function upgradeUser($userId)
    {
        $user = $this->tenant->createUser();
        $user->fillResource($userId);
        $user->upgrade();
    }

    /**
     * @param $mspGroupId
     * @param $endUserLogin
     * @return null|Service_MspApiClient_Response_Struct_UserItem
     * @throws Exception
     */
    private function findUserByLogin($mspGroupId, $endUserLogin)
    {
        $users = $this->getMspApi()->getUsers($mspGroupId);
        foreach ($users as $user) {
            if ($user->getLogin() === $endUserLogin) {
                return $user;
            }
        }
        return null;
    }

    /**
     * @return \Acronis\App\CloudApi\Api
     * @throws Exception
     */
    private function getCloudApi()
    {
        return $this->memoize(function () {
            $tenant = $this->tenant;
            $datacenter = $tenant->datacenterId ? $this->_getDatacenterById($tenant->global, $tenant->datacenterId) : $tenant->datacenter;

            return $this->createCloudApiInstance($datacenter);
        }, $this->tenant->tenantId);
    }

    /**
     * @param $user
     * @param $applicationId
     * @param $userQuota
     * @param $allCounters
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function obtainUserLimitsFromCloud($user, $applicationId, $userQuota, $allCounters)
    {
        $newOfferingItems = [];
        /** @var $user users */
        $userResource = new \users();
        $userResource->fillResource($user->aps->id);
        $offeringItemsFromCloud = $this->getCloudApi()->getTenantApplicationOfferingItems($userResource->lazyGetPersonalTenantId(), $applicationId);
        // This is dirty hack in order not to refactor logic which operates with not mutually ambiguous counters <-> offering items
        $offeringItemsToCounters = array_flip(array_filter($allCounters['offering_items'],
            function ($key) {
                return strpos($key, 'Hosted') === false;
            },
            ARRAY_FILTER_USE_KEY
        ));

        $counters = $allCounters['counters'];
        foreach ($offeringItemsFromCloud as $offeringItem) {
            $userOfferingItem = new UserOfferingItem();
            $userOfferingItem->name = $offeringItem->getName();
            if ($offeringItem->hasQuota() && $offeringItem->getQuota()->hasValue()) {
                $userOfferingItem->limit = $offeringItem->getQuota()->getValue();
            } elseif (!$userQuota) {
                $userOfferingItem->limit = self::UNLIMITED_VALUE;
            } else {
                $newUserCounterName = $offeringItemsToCounters[$userOfferingItem->name];

                $newCounterInfo = array_filter($counters, function ($value, $key) use ($newUserCounterName) {
                    if ($newUserCounterName === $key) {
                        return true;
                    }
                    foreach ($value as $v) {
                        if ($v['name'] === $newUserCounterName) {
                            return true;
                        }
                    }
                    return false;
                }, ARRAY_FILTER_USE_BOTH);

                $found = false;
                if ($newCounterInfo && !is_null($user->offeringItems)) {
                    $isForUser = false;
                    $stdCounterName = key($newCounterInfo);
                    $newCountersNames = array_map(function ($c) use (&$isForUser){
                        $isForUser |= Arr::get($c, 'isForUser', false);
                        return $c['name'];
                    }, reset($newCounterInfo));
                    $newCountersNames[] = $stdCounterName;
                    $isForUser |= $stdCounterName == 'LocalStorageUsage';
                    if (!$isForUser) {
                        $userOfferingItem->limit = self::UNLIMITED_VALUE;
                        continue;
                    }

                    foreach ($user->offeringItems as $oldOfferingItem) {
                        if (isset($offeringItemsToCounters[$oldOfferingItem->name])
                            && in_array($offeringItemsToCounters[$oldOfferingItem->name], $newCountersNames)) {
                            $found = true;
                            break;
                        }
                    }
                }
                $userOfferingItem->limit = $found ? self::UNLIMITED_VALUE : 0;
            }

            $newOfferingItems[] = $userOfferingItem;
        }

        $user->offeringItems = $newOfferingItems;
        $this->getApsc()->updateResource($user->resource);
    }
}
