<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\BackupApplication;
use Acronis\App\CloudApi\Api as CloudApi;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Model_PropertyConverters_UsersConverterHelper as UsersConverterHelper;
use Acronis\Cloud\Client\Model\Users\UserPut;

/**
 * Class Model_Resource_Tenant_Reseller
 */
class Model_Resource_Tenant_Reseller extends Model_Resource_Tenant
{
    use CountersManagerAwareTrait;
    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateCounterLimits();
        $this->checkMinimalCommitmentTiers();
        $this->linkMinimalCommitmentTier(false);

        $this->_verifyAccountDoesNotAlreadyHaveResellerSubscription();
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateStartProductionTime();

        $this->linkMinimalCommitmentTier(false);
        $tenant = $this->tenant;
        if (!$tenant->minimalCommitmentTier->isEmpty()) {
            $this->_applyMinimalCommitmentTierDiscountForAccount();
        }

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkMinimalCommitmentTier($realLink);

        $this->updateCounterLimits();
        $this->_retrieveUsageForReseller();
    }

    public function configure($new)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkMinimalCommitmentTier($realLink);

        $this->_updateAllGroupsForReseller();
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_cancelMinimalCommitmentTierDiscountForAccount();

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            Log::getInstance()->event(self::LOG_UNPROVISIONING_TENANT_IS_MIGRATION);
            return;
        }

        $apsApi = $this->getApsApi();
        $tenants = $apsApi->getTenantsByAccountId($tenant->accountId, $tenant->aps->id);
        if (count($tenants)) {
            return;
        }
        $groups = $this->_getListOfGroupIdForReseller();
        foreach ($groups as $group) {
            $datacenter = $this->_getDatacenterById($tenant->global, $group->datacenterId);
            $mspApi = $this->createMspApiInstance($datacenter);

            /**
             * @var Service_MspApiClient_Response_Struct_AdministratorItem[] $admins
             */
            $admins = $mspApi->getAdministrators($group->groupId);
            foreach ($admins as $admin) {
                $mspApi->deleteAdministrator($admin->getId(), $admin->getVersion());
            }
        }
        Log::getInstance()->event(self::LOG_WARNING_UNPROVISION_TENANT_IS_RESELLER);
    }

    public function enable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkMinimalCommitmentTier();
    }

    public function disable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        // Do nothing
    }

    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_array($tenant->storageRelations)) {
            /**
             * @var StorageRelation $storageRelation
             */
            foreach ($tenant->storageRelations as &$storageRelation) {
                if (!isset($storageRelation->datacenterId)) {
                    $storageRelation->datacenterId = globals::DEFAULT_DATACENTER_ID;
                }
            }
        }

        $this->updateStartProductionTime();
        $this->update();

        $tenant->global->reload();

        $realLink = !$tenant->subscription->disabled;
        $this->linkMinimalCommitmentTier($realLink);
        $needUpdate = $tenant->minimalCommitmentTier->isEmpty() ?
            $this->_cancelMinimalCommitmentTierDiscountForAccount() :
            $this->_applyMinimalCommitmentTierDiscountForAccount();
        if ($needUpdate) {
            $this->update();
        }

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    /**
     * Further use in dedicated command which should be called after package upgrade  from .spec
     */
    public function syncCloud()
    {
        $this->_updateAllGroupsForReseller();
    }

    /* ======= RESOURCE EVENTS ======= */

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkMinimalCommitmentTier();
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkMinimalCommitmentTier($realLink);

        $needUpdate = $tenant->minimalCommitmentTier->isEmpty() ?
            $this->_cancelMinimalCommitmentTierDiscountForAccount() :
            $this->_applyMinimalCommitmentTierDiscountForAccount();

        if ($this->updateStartProductionTime()) {
            $needUpdate = true;
        }
        if ($this->updateCounterLimits()) {
            $needUpdate = true;
        }
        if ($needUpdate) {
            $this->update();
        }
        $this->_updateAllGroupsForReseller();
        if (!$tenant->subscription->disabled) {
            $this->updateAllPrivilegesForReseller();
        }
    }

    public function onUserChanged($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateAllGroupsForReseller();
    }

    public function onAccountChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateAllGroupsForReseller();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @param $parameters
     * @param $datacenterId
     * @throws \Rest\RestException
     */
    public function createAdministrator($parameters, $tenantId = '', $datacenterId = '')
    {
        $tenant = $this->tenant;
        $this->validateDatacenterId($datacenterId);
        $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        $mspApi = $this->createMspApiInstance($datacenter);

        $resellerMapHelper = new Model_ResellerMapByDatacenterHelper($this->getApsc(), $tenant->global, $datacenter);
        $resellerGroupId = $resellerMapHelper->getResellerIdFromMap($tenant->accountId);
        //create Reseller in ABC if not exists
        if (is_null($resellerGroupId)) {
            //get Root Reseller and try to provision group in ABC
            $parentGroupId = $this->_getOrCreateParentGroupId($datacenter);
            $mspGroupManager = $this->createMspGroupManagerInstance($mspApi, $this->getApsApi());
            $resellerGroupId = $mspGroupManager->createGroupForReseller($tenant, $parentGroupId, $datacenter->id, $this);

            $resellerMapHelper->addResellerIdToMap($tenant->accountId, $resellerGroupId);
            $resellerMapHelper->updateResellerMap();
        }

        // Small note: to create read-only admin we use APIv1. In order to minify changes
        // Service_MspApiClient_Response_Struct_CreateGroup is not extended with email and login
        // fields. So $parameters->email  $parameters->email
        $adminInfo = $this->createAdministratorApiV1($parameters, $resellerGroupId, $datacenterId);

        try {
            $mspApi->postMailActivate(new Service_MspApiClient_Request_Struct_PostMailActivate(
                $parameters->email, $parameters->login
            ));
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }

        // This method is throwing no content exception
        $this->updateAdministrator($adminInfo->getUuid(), $parameters, $datacenterId);
    }

    /**
     * @deprecated
     * @param object $parameters
     * @param string $groupId
     * @param string $datacenterId
     * @return \Service_MspApiClient_Response_Struct_CreateGroup
     * @throws \Rest\RestException
     */
    public function createAdministratorApiV1($parameters, $groupId = '', $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        //create Administrator in Reseller's group in ABC
        $adminInfo = new Service_MspApiClient_Request_Struct_AdministratorInfo();
        $adminInfo->setType(self::RESELLER_GROUP_ADMIN_ACCOUNT_PERMS);

        $adminInfo->setLogin($parameters->login);
        $adminInfo->setEmail($parameters->email);

        if (isset($parameters->firstName)) {
            $adminInfo->setFirstName($parameters->firstName);
        }
        if (isset($parameters->lastName)) {
            $adminInfo->setLastName($parameters->lastName);
        }

        $tenant = $this->tenant;
        $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        $mspApi = $this->createMspApiInstance($datacenter);

        return $mspApi->createAdministrator($adminInfo, $groupId);
    }

    public function updateAdministrator($adminId, $parameters, $datacenterId = '')
    {
        $this->validateDatacenterId($datacenterId);
        parent::updateAdministrator($adminId, $parameters, $datacenterId);
    }

    /**
     * @return array
     * @throws \Rest\RestException
     */
    public function getAdministrators()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $list = array();
        $groups = $this->_getListOfGroupIdForReseller();
        foreach ($groups as $group) {
            $datacenter = $this->_getDatacenterById($tenant->global, $group->datacenterId);
            $cloudApi = $this->createCloudApiInstance($datacenter);
            $mspApi = $this->createMspApiInstance($datacenter);

            $resellerTenantId = $mspApi->getGroupById($group->groupId)->getUuid();
            $admins = $cloudApi->getTenantUsers($resellerTenantId);
            foreach ($admins as $adminId) {
                $admin = $cloudApi->getUser($adminId);
                $item = new stdClass();
                $item->datacenterId = $datacenter->id;
                $item->datacenterName = $datacenter->name;
                $item->id = $admin->getId();
                $item->firstName = $admin->getContact()->getFirstName();
                $item->lastName = $admin->getContact()->getLastName();
                $item->login = $admin->getLogin();
                $item->email = $admin->getContact()->getEmail();
                $item->status = UsersConverterHelper::getUserStatus($admin->getEnabled(), $admin->getActivated());
                $list[] = $item;
            }
        }
        return $list;
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @return object
     */
    public function getAdministrator($adminId, $datacenterId = '')
    {
        $this->validateDatacenterId($datacenterId);
        return parent::getAdministrator($adminId, $datacenterId);
    }

    /**
     * @param $adminId
     * @param $datacenterId
     * @throws \Rest\RestException
     */
    public function deleteAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->validateDatacenterId($datacenterId);

        $tenant = $this->tenant;
        $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        $mspApi = $this->createMspApiInstance($datacenter);

        $admin = $mspApi->getAdministratorById($adminId);
        $mspApi->deleteAdministrator($admin->getId(), $admin->getVersion());

        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param $datacenterId
     * @return stdClass
     * @throws \Rest\RestException
     */
    public function getAdministratorConsole($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->validateDatacenterId($datacenterId);

        $tenant = $this->tenant;
        $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        $mspApi = $this->createMspApiInstance($datacenter);

        $admin = $mspApi->getAdministratorById($adminId);
        $brand = $mspApi->getBrand($admin->getGroup());
        $url = $brand->getAccountServerUrl();
        $token = $mspApi->getImpersonationToken($adminId);

        $console = new stdClass();
        $console->url = sprintf(self::URL_REDIRECT_MASK, $url, $token);
        return $console;
    }

    /**
     * @param $adminId
     * @param $datacenterId
     * @throws \Rest\RestException
     */
    public function resendAdministratorLink($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->validateDatacenterId($datacenterId);

        $tenant = $this->tenant;
        $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        $mspApi = $this->createMspApiInstance($datacenter);

        $admin = $mspApi->getAdministratorById($adminId);
        $mspApi->postMailActivate(
            new Service_MspApiClient_Request_Struct_PostMailActivate($admin->getEmail(), $admin->getLogin())
        );

        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param $datacenterId
     * @throws \Rest\RestException
     */
    public function enableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));


        $this->validateDatacenterId($datacenterId);
        $datacenter = $this->_getDatacenterById($this->tenant->global, $datacenterId);
        $cloudApi = $this->createCloudApiInstance($datacenter);
        $user = $cloudApi->getUser($adminId);
        if ($user->getEnabled()) {
            $this->throwNoContent();
        }
        $userUpdate = new UserPut();
        $userUpdate->setVersion($user->getVersion());
        $userUpdate->setEnabled(true);
        $cloudApi->updateUser($adminId, $userUpdate);
        $this->throwNoContent();
    }

    public function adminUrlRedirect()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    /**
     * @param $adminId
     * @param $datacenterId
     * @throws \Rest\RestException
     */
    public function disableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->validateDatacenterId($datacenterId);
        $datacenter = $this->_getDatacenterById($this->tenant->global, $datacenterId);
        $cloudApi = $this->createCloudApiInstance($datacenter);
        $user = $cloudApi->getUser($adminId);

        if (!$user->getEnabled()) {
            $this->throwNoContent();
        }

        $userUpdate = new UserPut();
        $userUpdate->setVersion($user->getVersion());
        $userUpdate->setEnabled(false);
        $cloudApi->updateUser($adminId, $userUpdate);
        $this->throwNoContent();
    }

    /**
     * @return array
     */
    public function getDatacenters()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $datacenters = array();
        $tenant = $this->tenant;
        foreach ($tenant->global->datacenters as $datacenter) {
            $datacenters[] = (object)$datacenter->toArray();
        }
        return $datacenters;
    }

    /**
     * @return array
     * @throws Exception
     */
    public function getStorages()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        return $this->memoize(function () {
            $response = [];

            $tenant = $this->tenant;
            $resellerRelations = $tenant->global->resellerRelations;
            if (!is_array($resellerRelations)) {
                return $response;
            }

            foreach ($resellerRelations as $relation) {
                if ($relation->accountId != $tenant->accountId) {
                    continue;
                }

                $datacenter = $this->_getDatacenterById($tenant->global, $relation->datacenterId);
                $mspApi = $this->createMspApiInstance($datacenter);
                $groupId = $relation->groupId;

                $storages = $mspApi->getStorages($groupId);
                if (is_null($storages)) {
                    Log::getInstance()->error(static::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
                    throw new Exception(static::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
                }

                foreach ($storages as $storage) {
                    /** @var Service_MspApiClient_Response_Struct_Storage $storage */
                    if ($storage->getInherited()) {
                        continue;
                    }
                    $response[] = [
                        static::FIELD_STORAGE_ID => (string)$storage->getId(),
                        static::FIELD_STORAGE_NAME => $storage->getName(),
                        static::FIELD_STORAGE_DATACENTER_ID => $datacenter->id,
                        static::FIELD_STORAGE_DATACENTER_NAME => $datacenter->name,
                    ];
                }
            }

            return $response;
        });
    }

    /**
     * @return array
     */
    public function getServicePlans()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $profiles = $tenant->global->storages;
        if (!count($profiles)) {
            return array();
        }

        $templates = $this->getApsApi()->getServiceTemplatesWithLimits();
        if (!count($templates)) {
            return array();
        }

        $plans = $this->getPbaApi()->getServicePlans($tenant->accountId);
        if (!count($plans)) {
            return array();
        }

        $servicePlans = array();
        foreach ($plans as $plan) {
            $formattedPlan = $this->_formatServicePlan($plan, $templates, $profiles);
            if (is_null($formattedPlan)) {
                continue;
            }
            $servicePlans[] = $formattedPlan;
        }
        return $servicePlans;
    }

    /**
     * @param $planId
     * @return object|null
     * @throws \Rest\RestException
     */
    public function getServicePlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $profiles = $tenant->global->storages;
        if (!count($profiles)) {
            $this->throwNoContent();
        }

        $templates = $this->getApsApi()->getServiceTemplatesWithLimits();
        if (!count($templates)) {
            $this->throwNoContent();
        }

        $plan = $this->getPbaApi()->getServicePlan($tenant->accountId, $planId);
        if (is_null($plan)) {
            $this->throwNoContent();
        }

        $formattedPlan = $this->_formatServicePlan($plan, $templates, $profiles);
        if (is_null($formattedPlan)) {
            $this->throwNoContent();
        }
        return $formattedPlan;
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     */
    public function loginExists($login, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        return $this->checkLoginExist($login, $datacenterId);
    }

    /**
     * @param $planId
     * @param $parameters
     * @throws Exception
     * @throws \Rest\RestException
     */
    public function setStorageForPlan($planId, $parameters)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        if (!isset($parameters->storageId)) {
            throw new Exception(self::EXCEPTION_STORAGE_ID_MUST_BE_PRESENT_MESSAGE);
        }
        if (!isset($parameters->datacenterId)) {
            throw new Exception(self::EXCEPTION_DATACENTER_ID_MUST_BE_PRESENT_MESSAGE);
        }

        $storageId = $parameters->storageId;
        $datacenterId = $parameters->datacenterId;
        Log::getInstance()->event(sprintf(self::LOG_CALLED_SET_STORAGE_FOR_PLAN, $storageId, $planId));
        $tenant = $this->tenant;
        if (!is_array($tenant->storageRelations)) {
            $tenant->storageRelations = array();
        }
        foreach ($tenant->storageRelations as &$item) {
            if ($item->planId == $planId) {
                $item->storageId = $storageId;
                $item->datacenterId = $datacenterId;
                $this->update();
                $this->throwNoContent();
            }
        }
        $relation = new StorageRelation();
        $relation->planId = $planId;
        $relation->storageId = $storageId;
        $relation->datacenterId = $datacenterId;
        $tenant->storageRelations[] = $relation;
        $this->update();
        $this->throwNoContent();
    }

    /**
     * @param $planId
     * @throws \Rest\RestException
     */
    public function unsetStorageForPlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (!is_array($tenant->storageRelations)) {
            $tenant->storageRelations = array();
        }
        foreach ($tenant->storageRelations as $key => $item) {
            if ($item->planId == $planId) {
                unset($tenant->storageRelations[$key]);
                $this->update();
            }
        }
        $this->throwNoContent();
    }

    /**
     * @return array
     */
    public function updateUsageAndLimits()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->updateCounterLimits();
        $tenantUsages = $this->fetchTenantLimitsAndUsages();

        return [
            static::TENANT => $tenantUsages,
            static::USER_USAGE => [],
            static::ERROR => null,
        ];
    }

    public function getDatacenter()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getQuotasStatistics()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getUserGuideUrl(){
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    /* ======= OTHER METHODS ======= */

    public function resolveDatacenterId()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->throwNotApplicableMethodException();
    }

    public function resolveBackupStorageId()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->throwNotApplicableMethodException();
    }

    private function _verifyAccountDoesNotAlreadyHaveResellerSubscription()
    {
        $tenant = $this->tenant;
        $resellerTenants = $this->getApsApi()->getResellersByAccountId($tenant->account->id);
        if (count($resellerTenants) > 0) {
            throw new \Rest\RestException(409, self::EXCEPTION_ACCOUNT_ALREADY_HAS_RESELLER_SUBSCRIPTION);
        }
    }

    /**
     * @return bool need update resource
     * @throws \Rest\RestException
     */
    private function _applyMinimalCommitmentTierDiscountForAccount()
    {
        $tenant = $this->tenant;
        if (!empty($tenant->minimalCommitmentDiscountId)) {
            return false;
        }
        $accountId = $tenant->account->id;
        $discountId = $this->_getMinimalCommitmentDiscountForAccount($accountId);

        try {
            $this->getPbaApi()->addAccountInDiscount($accountId, $discountId);
        } catch (Service_PbaApi_Api_Exception $e) {
            if (!$e->isDiscountForAccountAlreadyExistsException()) {
                throw new \Rest\RestException(409, $e->getMessage() . ' ' . $e->getPrevious()->getMessage());
            }
        }
        $this->addAccountToProcess($accountId, self::ACCOUNT_PROCESS_TYPE_ONBOARDING);
        $tenant->minimalCommitmentDiscountId = $discountId;
        return true;
    }

    /**
     * @param $accountId
     * @return int
     * @throws Exception
     */
    private function _getMinimalCommitmentDiscountForAccount($accountId)
    {
        $accountInfo = $this->getOpenApi()->getAccountInfo($accountId);
        $parentAccountId = $accountInfo->getParentAccountId();

        $tenant = $this->tenant;
        $discountRelations = $tenant->global->discountRelations;
        if (empty($discountRelations)) {
            throw new \Rest\RestException(409, Str::format(self::EXCEPTION_NO_DISCOUNT_FOR_ACCOUNT, $parentAccountId));
        }

        $relations = array_filter($discountRelations, function ($relation) use ($parentAccountId) {
            return $relation->accountId == $parentAccountId;
        });

        if (count($relations) == 0) {
            throw new \Rest\RestException(409, Str::format(self::EXCEPTION_NO_DISCOUNT_FOR_ACCOUNT, $parentAccountId));
        }

        if (count($relations) > 1) {
            throw new \Rest\RestException(409, Str::format(self::EXCEPTION_SEVERAL_DISCOUNTS_FOR_ACCOUNT, $parentAccountId));
        }

        $relation = reset($relations);
        $discountId = $relation->discountId;
        return $discountId;
    }

    /**
     * @param $accountId
     * @param $type
     */
    private function addAccountToProcess($accountId, $type)
    {
        $date = time();

        $tenant = $this->tenant;
        $global = $this->getApsc()->getResource($tenant->global->aps->id);
        $global->accountToProcess = $global->accountToProcess ?: array();
        usort($global->accountToProcess, function ($record1, $record2) {
            return $record1->date - $record2->date;
        });
        $accounts = array_filter($global->accountToProcess, function ($account) use ($accountId) {
            return $account->accountId == $accountId;
        });
        $lastRecord = end($accounts);
        if ($lastRecord && $lastRecord->type === $type) {
            return;
        }

        $apsApi = $this->getApsApi();

        $children = $apsApi->getChildAccounts($tenant->account);
        $subscriptions = array();
        foreach ($children as $account) {
            $accountSubscriptions = $apsApi->getAccountSubscriptions($account);
            $subscriptionsIds = array_map(function ($subscription) {
                return $subscription->subscriptionId;
            }, $accountSubscriptions);
            $subscriptions = array_merge($subscriptions, $subscriptionsIds);
        }

        $accountToProcess = new AccountToProcess();
        $accountToProcess->accountId = $accountId;
        $accountToProcess->date = $date;
        $accountToProcess->type = $type;
        $accountToProcess->subscriptions = $subscriptions;

        $global->accountToProcess[] = $accountToProcess;
        $this->getApsc()->updateResource($global);
        $tenant->global->reload();
    }

    private function _updateAllGroupsForReseller()
    {
        $tenant = $this->tenant;
        $resellerRelations = $tenant->global->resellerRelations;
        if (!is_array($resellerRelations)) {
            return;
        }

        foreach ($resellerRelations as $relation) {
            if ($relation->accountId != $tenant->accountId) {
                continue;
            }

            $this->_updateGroupForReseller($relation->datacenterId, $relation->groupId);
        }
    }

    private function updateAllPrivilegesForReseller()
    {
        $tenant = $this->tenant;
        $resellerRelations = $tenant->global->resellerRelations;
        if (!is_array($resellerRelations)) {
            return;
        }

        foreach ($resellerRelations as $relation) {
            /** @var ResellerRelation $relation */
            if ($relation->accountId != $tenant->accountId) {
                continue;
            }

            $datacenter = $this->_getDatacenterById($tenant->global, $relation->datacenterId);
            $mspGroupManager = $this->createMspGroupManagerInstance(
                $this->createMspApiInstance($datacenter),
                $this->getApsApi()
            );
            $mspGroupManager->updatePrivilegesForCustomers($tenant, $this);
        }
    }

    /**
     * @param string $datacenterId
     * @param int $groupId
     * @throws Exception
     */
    private function _updateGroupForReseller($datacenterId, $groupId)
    {
        $tenant = $this->tenant;
        $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        $mspGroupManager = $this->createMspGroupManagerInstance(
            $this->createMspApiInstance($datacenter),
            $this->getApsApi()
        );
        $mspGroupManager->updateGroupForReseller($groupId, $tenant);
    }

    /**
     * @return bool need update resource
     * @throws \Rest\RestException
     */
    private function _cancelMinimalCommitmentTierDiscountForAccount()
    {
        $tenant = $this->tenant;
        if (empty($tenant->minimalCommitmentDiscountId)) {
            return false;
        }
        $accountId = $tenant->account->id;
        $discountId = $tenant->minimalCommitmentDiscountId;

        try {
            $this->getPbaApi()->moveToArchiveAccountInDiscount($accountId, $discountId);
        } catch (Service_PbaApi_Api_Exception $e) {
            throw new \Rest\RestException(409, $e->getMessage() . ' ' . $e->getPrevious()->getMessage());
        }
        $this->addAccountToProcess($accountId, self::ACCOUNT_PROCESS_TYPE_CANCELLATION);
        $tenant->minimalCommitmentDiscountId = null;
        return true;
    }

    /**
     * Return a list of groups for a reseller.
     * Each item has fields groupId and datacenterId.
     * @return array
     */
    private function _getListOfGroupIdForReseller()
    {
        $tenant = $this->tenant;
        $list = array();
        if (!is_array($tenant->global->resellerRelations)) {
            return $list;
        }
        foreach ($tenant->global->resellerRelations as $reseller) {
            if ($reseller->accountId == $tenant->accountId) {
                $item = new stdClass();
                $item->groupId = $reseller->groupId;
                $item->datacenterId = $reseller->datacenterId;
                $list[] = $item;
            }
        }
        return $list;
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_PlanInterface $plan
     * @param array $templates
     * @param array $profiles
     * @return null | object
     */
    private function _formatServicePlan($plan, $templates, $profiles)
    {
        $tenant = $this->tenant;

        $serviceTemplateId = $plan->getServiceTemplateId();
        if (!isset($templates[$serviceTemplateId])) {
            return NULL;
        }
        $template = $templates[$serviceTemplateId];

        $profile = $this->_getProfileForTemplate($template, $profiles);
        if (is_null($profile)) {
            return NULL;
        }

        $planId = $plan->getPlanId();
        $storagePolicy = is_null($profile->storagePolicy) ? profile::STORAGE_POLICY_PARENT : $profile->storagePolicy;
        $storageId = NULL;
        $storageName = NULL;
        $datacenterId = NULL;
        $datacenterName = NULL;
        if ($storagePolicy == profile::STORAGE_POLICY_PARENT) {
            $storageId = $profile->storageId === self::STORAGE_DEFAULT ? self::STORAGE_DEFAULT : self::STORAGE_PROVIDER;
        } elseif ($storagePolicy == profile::STORAGE_POLICY_OWN) {
            $storage = $this->_getStorageForPlan($tenant, $planId);
            if (!is_null($storage)) {
                $storageId = $storage[self::FIELD_STORAGE_ID];
                $storageName = $storage[self::FIELD_STORAGE_NAME];
                $datacenterId = $storage[self::FIELD_STORAGE_DATACENTER_ID];
                $datacenterName = $storage[self::FIELD_STORAGE_DATACENTER_NAME];
            }
        }

        $servicePlan = new stdClass();
        $servicePlan->planId = $planId;
        $servicePlan->name = $plan->getName();
        $servicePlan->storageId = $storageId;
        $servicePlan->storageName = $storageName;
        $servicePlan->storagePolicy = $storagePolicy;
        $servicePlan->datacenterId = $datacenterId;
        $servicePlan->datacenterName = $datacenterName;

        return $servicePlan;
    }

    /**
     * @param \APS\ResourceBase $template
     * @param profile[] $profiles
     * @return null
     */
    private function _getProfileForTemplate($template, $profiles)
    {
        foreach ($profiles as $profile) {
            $resourceTypeId = $profile->resourceTypeId;
            if (isset($template->limits[$resourceTypeId])) {
                return $profile;
            }
        }
        return NULL;
    }

    /**
     * @throws \Acronis\App\UsageReport\Exception\ManagerException
     * @throws \Rest\RestException
     * @throws Exception
     */
    private function _retrieveUsageForReseller()
    {
        $tenant = $this->tenant;
        foreach ($this->getCountersManager()->getApplicationCounters(CloudApi::APPLICATION_TYPE_BACKUP) as $counter) {
            if ($counter->isFake()) {
                continue;
            }
            $tenant->{$counter->getName()}->usage = 0;
        }

        $resellerRelations = $tenant->global->resellerRelations;
        if (!is_array($resellerRelations)) {
            return;
        }

        foreach ($resellerRelations as $relation) {
            if ($relation->accountId != $tenant->accountId) {
                continue;
            }

            $datacenter = $this->_getDatacenterById($tenant->global, $relation->datacenterId);
            $this->updateCountersUsage($datacenter, $relation->groupId);
        }
    }

    protected function fetchTenantLimitsAndUsages()
    {
        // it's w/a to keep an existing architecture and do not refactor code till 3.0
        // use any datacenter and null fot resellerTenantId
        // due limits and usages will be fetched from subscription resources
        $resellerTenantId = null;
        $datacenters = $this->tenant->global->datacenters;
        $datacenter = count($datacenters) ? $datacenters[0] : null;
        if (is_null($datacenter)) {
            return [];
        }
        $mspApi = $this->createMspApiInstance($datacenter);

        $backupApplicationManager = new BackupApplication($this->tenant, $mspApi, $resellerTenantId);

        return $backupApplicationManager->fetchLimitsAndUsages();
    }
}
