<?php

use Acronis\App\CloudApi\Api;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\Cloud\Client\Model\AccessPolicies\AccessPolicy;
use Acronis\Cloud\Client\Model\AccessPolicies\AccessPolicyTrusteeType;
use Acronis\Cloud\Client\Model\Applications\Application;
use Acronis\Cloud\Client\Model\Common\Types\Role;
use Acronis\Cloud\Client\Model\OfferingItems\OfferingItemOutput;
use Acronis\Cloud\Client\Model\Users\User;
use Acronis\Cloud\Client\Model\Users\UserPost;
use http\Exception\InvalidArgumentException;
use Rest\RestException;

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Model_UserManager implements Model_UserManagerInterface
{
    use CountersManagerAwareTrait,
        MemoizeTrait;

    /** @var Service_MspApiClient_ApiInterface */
    private $_mspApi;
    /** @var FacadeInterface */
    private $_facade;
    /** @var Service_MspApiClient_Request_Struct_UpdateUserInfo */
    private $_updateUserInfo;

    /**
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param FacadeInterface $facade
     * @param Service_MspApiClient_Request_Struct_UpdateUserInfo $updateUserInfo
     */
    public function __construct(Service_MspApiClient_ApiInterface $mspApi,
        FacadeInterface $facade = NULL,
        Service_MspApiClient_Request_Struct_UpdateUserInfo $updateUserInfo = NULL)
    {
        $this->_mspApi= $mspApi;
        $this->_facade = is_null($facade) ? new Facade() : $facade;
        $this->_updateUserInfo = is_null($updateUserInfo) ? new Service_MspApiClient_Request_Struct_UpdateUserInfo() : $updateUserInfo;
    }

    /**
     * @param string $login
     * @param string $password
     */
    public function activateUser($login, $password)
    {
        $token = $this->_mspApi->mailActivate($login);
        $this->_mspApi->activateAccountWithPassword($token, $password);
    }

    /**
     * @param string $email
     * @param string $password
     * @throws Exception
     */
    public function resetUserPassword($email, $password)
    {
        try
        {
            $token = $this->_mspApi->mailReset($email);
            $this->_mspApi->resetPassword($token, $password);
        }
        catch (Exception $e)
        {
            if ($e->getCode() == Service_BaseClient_RestTransport::HTTP_CODE_FORBIDDEN)
            {
                return;
            }
            throw $e;
        }
    }

    /**
     * @param int $userId
     * @param int $version
     * @return int
     */
    public function enableUser($userId, $version)
    {
        return $this->_setUserStatus($userId, $version, Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_ENABLE);
    }

    /**
     * @param int $userId
     * @param int $version
     * @return int
     */
    public function disableUser($userId, $version)
    {
        return $this->_setUserStatus($userId, $version, Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE);
    }

    /**
     * @param int $groupId
     * @return array
     */
    public function getBackItems($groupId)
    {
        $backupConsole = $this->_mspApi->getBackupConsole($groupId);
        $link = $backupConsole->getHost();
        $accessToken = $backupConsole->getToken();

        $amsApi = $this->_facade->getAmsApi($link);
        $amsApi->createRemoteConnection($accessToken);
        return $amsApi->getBackupItems();
    }

    /**
     * @param string $email
     * @param int $groupId
     * @return NULL|Service_MspApiClient_Response_Struct_UserInfo
     */
    public function getUserByEmail($email, $groupId)
    {
        $users = $this->_mspApi->getUsers($groupId);
        if (!is_array($users))
        {
            return NULL;
        }
        foreach ($users as $userItem)
        {
            if ($userItem->getEmail() == $email)
            {
                return $this->_mspApi->getUserById($userItem->getId());
            }
        }

        return NULL;
    }

    /**
     * @param $userId
     * @param bool $force
     * @return Service_MspApiClient_Response_Struct_UserInfo|User
     */
    public function getUserById($userId, $force = false)
    {
        return $this->memoize(function() use ($userId) {
            if (!$userId) {
                return null;
            }

            return is_numeric($userId) ?
                $this->_mspApi->getUserById($userId) :
                $this->createCloudApi()->getUser($userId);
        }, $userId, $force);
    }

    /**
     * todo remove this w/a after fix of ci-lib-php-cloud-client (array of one element => single string)
     * @param $user
     * @return string
     */
    public function getUserPersonalTenantId($user)
    {
        if (!$user) {
            throw new InvalidArgumentException('Cannot get personal tenant id, no user provided');
        }

        $personalTenantId = $user->getPersonalTenantId();

        if (is_array($personalTenantId)) {
            return reset($personalTenantId);
        }

        if (is_string($personalTenantId)) {
            return $personalTenantId;
        }

        throw new InvalidArgumentException('User\'s personal tenant id is not in valid format');
    }

    /**
     * @param string $companyName
     * @return NULL|Service_MspApiClient_Response_Struct_GroupInfo
     */
    public function getGroupByName($companyName)
    {
        $groups = $this->_mspApi->search($companyName);
        if (!is_array($groups))
        {
            return NULL;
        }
        foreach ($groups as $group)
        {
            if ($group->getName() == $companyName)
            {
                return $group;
            }
        }
        return NULL;
    }

    /**
     * @param int $userId
     * @param int $version
     * @param int $status
     * @return int
     */
    private function _setUserStatus($userId, $version, $status)
    {
        $this->_updateUserInfo->setStatus($status);
        return $this->_mspApi->updateUser($userId, $this->_updateUserInfo, $version);
    }

    /**
     * add administrator account to baas with data from PA staff member
     *
     * @param int $mspGroupId
     * @param int $subscriptionId
     * @param UserPost $userInfo
     * @throws Exception
     */
    public function provisionAdministratorAccount($mspGroupId, $subscriptionId, UserPost $userInfo)
    {
        try {
            $cloudApi = $this->createCloudApi();
            $createdAdmin = $cloudApi->createUser($userInfo);
            $adminAP = new AccessPolicy();
            $adminAP->setId(ApiInterface::NULL_UUID)
                ->setIssuerId(ApiInterface::NULL_UUID)
                ->setTrusteeId($createdAdmin->getId())
                ->setTrusteeType(AccessPolicyTrusteeType::USER)
                ->setVersion(0)
                ->setTenantId($userInfo->getTenantId())
                ->setRoleId(Role::PARTNER_ADMIN);
            $cloudApi->updateUserAccessPolicies($createdAdmin->getId(), [$adminAP]);
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_CONFLICT) {
                throw $e;
            }
            $admin = $this->getCurrentAdministrator($mspGroupId, $userInfo);
            if (is_null($admin)) {
                throw $e;
            }
            if ($admin->getStatus() !== Service_MspApiClient_Response_Struct_UserInfo::STATUS_INACTIVE) {
                Log::getInstance()->event(
                    sprintf(
                        users::LOG_FORMAT_ATTEMPT_TO_CREATE_EXISTING_MSP_USER,
                        $userInfo->getLogin(),
                        $subscriptionId
                    )
                );
                throw new Exception(users::EXCEPTION_USER_ALREADY_EXISTS);
            }
        }
        $this->sendActivateEmailForUser($userInfo);
    }

    /**
     * @param users $user
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function updateUserPersonalQuotas($user)
    {
        $personalTenantId = $this->initUserPersonalTenantId($user);
        if (!$personalTenantId || !count($user->offeringItems)) {
            return;
        }

        $cloudApi = $this->createCloudApi();
        $application = $this->getApplication(ApiInterface::APPLICATION_TYPE_BACKUP);
        $applicationId = $application->getId();

        $offeringItemsToUpdate = [];
        $availableOfferingItemsHashmap = $cloudApi->getOfferingItemsAvailableForChildHashMap($personalTenantId);
        $userOfferingItemsHashmap = Arr::map($user->offeringItems, 'name');
        $userApplicableOfferingItemsSet = $this->getCountersManager()->getUserApplicableOfferingItemsSet();

        foreach ($availableOfferingItemsHashmap as $offeringItemName => $offeringItemFromCloud) {
            /** @var OfferingItemOutput $offeringItemFromCloud */

            if ($offeringItemFromCloud->getStatus() === 0) {
                // In this case we need to exclude this OI from user OI's list and skip it
                $user->offeringItems = array_filter($user->offeringItems, function($userOi) use ($offeringItemName) {
                    /**  @var UserOfferingItem $userOi */
                    return $userOi->name !== $offeringItemName;
                });
                continue;
            }
            $offeringItemFromUser = Arr::get($userOfferingItemsHashmap, $offeringItemName);

            // We update $offeringItemFromCloud based on offering item from user limit
            if ($offeringItemFromUser && isset($userApplicableOfferingItemsSet[$offeringItemName])) {
                $limit = Arr::get($offeringItemFromUser, 'limit');
                $quota = $offeringItemFromCloud->getQuota();
                if ($limit !== Model_Resource_Tenant::UNLIMITED_VALUE) {
                    $quota->setValue($limit);
                    $quota->setOverage(0);
                } else {
                    $quota->setValue(null);
                    $quota->setOverage(null);
                }
                $offeringItemFromCloud->setQuota($quota);
            } else {
                $offeringItemFromCloud->unsetQuota();
            }

            // And use updated (or not) $offeringItemFromCloud for further use in PUT request
            $offeringItemsToUpdate[] = $offeringItemFromCloud;
        }

        $cloudApi->updateApplicationOfferingItemsForTenantsChain(
            $user->tenant->tenantId, $personalTenantId, $applicationId, $offeringItemsToUpdate
        );
    }

    protected function initUserPersonalTenantId($user)
    {
        if (!$user->personalTenantId) {
            $userId = !empty($user->userId) ? $user->userId : $user->mspUserId;
            if (is_numeric($userId)) {
                $userId = $this->getUserById($userId)->getUuid();
            }
            $cloudUser = $this->getUserById($userId);
            $user->personalTenantId = $this->getUserPersonalTenantId($cloudUser);
        }

        return $user->personalTenantId;
    }

    /**
     * @param $applicationType
     * @return Application|null
     */
    protected function getApplication($applicationType)
    {
        return $this->memoize(function() use ($applicationType) {
            $cloudApi = $this->createCloudApi();

            return $cloudApi->getApplication($applicationType);
        }, $applicationType);
    }

    /**
     * @return Api
     */
    protected function createCloudApi()
    {
        return $this->memoize(function () {
            return new Api(
                $this->_mspApi->getUrl(),
                $this->_mspApi->getLogin(),
                $this->_mspApi->getPassword()
            );
        });
    }

    /**
     * @param $mspGroupId
     * @param UserPost $userInfo
     * @return null|Service_MspApiClient_Response_Struct_AdministratorItem
     */
    private function getCurrentAdministrator($mspGroupId, UserPost $userInfo)
    {
        $administrators = $this->_mspApi->getAdministrators($mspGroupId);
        if (empty($administrators)) {
            return null;
        }
        foreach ($administrators as $administrator) {
            /** @var Service_MspApiClient_Response_Struct_AdministratorItem $administrator */
            if ($administrator->getLogin() == $userInfo->getLogin()) {
                return $administrator;
            }
        }
        return null;
    }

    /**
     * @param UserPost $userInfo
     */
    private function sendActivateEmailForUser(UserPost $userInfo)
    {
        $this->_mspApi->postMailActivate(new Service_MspApiClient_Request_Struct_PostMailActivate(
            $userInfo->getContact()->getEmail(),
            $userInfo->getLogin()
        ));
    }
}
