<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Services\CountersManager;

use Acronis\App\Base\ApplicationCounter;
use Acronis\App\Base\ApplicationCounter as Counter;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Utils\UomConverter as Uom;
use Arr;

class CountersManager
{
    // This is temporary solution, should be removed in future
    const PREFIXES_COUNTERS = ['Adv', 'Dre', 'PAdv', 'PDre', 'P'];
    const PREFIXES_OFFERING_ITEMS = ['adv_', 'dre_', 'p_adv_', 'p_dre_', 'p_'];
    const NORMALIZATION_MAP = [
        'adv_o365_seats' => 'mailboxes',
        'dre_o365_seats' => 'mailboxes',
        'p_dre_o365_seats' => 'mailboxes',
        'p_adv_o365_seats' => 'mailboxes',
        'p_o365_seats' => 'mailboxes'
    ];

    /** @var array */
    private $applications;

    /** @var Counter[] */
    private $countersHash;

    /** @var Counter[] */
    private $offeringItemsHash;

    /**
     * Counters constructor.
     *
     * @param array $countersData
     */
    public function __construct(array $countersData)
    {
        $this->countersHash = [];
        $this->offeringItemsHash = [];
        $this->applications = [];

        foreach ($countersData as $counterData) {
            $counter = new Counter($counterData);

            $this->countersHash[$counter->getName()] = $counter;
            $this->offeringItemsHash[$counter->getOfferingItemName()] = $counter;
            foreach ($counter->getPartnerOfferingItems() as $partnerOfferingItem) {
                $partnerCounter = clone $counter;
                $partnerCounter->setTypesPartnerAndReseller();
                $this->offeringItemsHash[$partnerOfferingItem] = $partnerCounter;
            }
            $this->applications[$counter->getApplicationType()][] = $counter;
        }
    }

    public function getUserApplicableOfferingItemsSet()
    {
        return array_reduce($this->getAllCounters(), function ($set, $counter) {
            /** @var ApplicationCounter $counter */
            if ($counter->isForUser()) {
                $set[$counter->getOfferingItemName()] = true;
            }

            return $set;
        }, []);
    }

    /**
     * @param string $counterName
     * @return Counter|null
     */
    public function getCounterByName($counterName)
    {
        return Arr::get($this->countersHash, $counterName, null);
    }

    /**
     * @param string $offeringItemName
     * @return Counter|null
     */
    public function getCounterByOfferingItemName($offeringItemName)
    {
        return Arr::get($this->offeringItemsHash, $offeringItemName, null);
    }

    /**
     * @param string[] $offeringItemsNames
     * @param callable $skipCondition
     * @return Counter[]
     */
    public function findCountersByOfferingItemNames($offeringItemsNames, $skipCondition = null)
    {
        return array_values(array_filter($this->countersHash,
            function ($counter) use ($offeringItemsNames, $skipCondition) {
                /** @var  ApplicationCounter $counter */
                return (!$skipCondition || $skipCondition($counter)) && in_array($counter->getOfferingItemName(),
                        $offeringItemsNames);
            }));
    }

    /**
     * @param string $applicationType
     * @return Counter[]
     */
    public function getApplicationCounters($applicationType)
    {
        return Arr::get($this->applications, $applicationType, []);
    }

    /**
     * @return Counter[]
     */
    public function getAllCounters()
    {
        return array_values($this->countersHash);
    }

    /**
     * @return array
     */
    public function getAllCountersMap()
    {
        $applicationCounters = require DIR_APP . DS . 'application_counters.php';
        $stdCounters = array_filter($applicationCounters, function ($counter) {
            return !isset($counter['aggregation_rules']['edition']) || ($counter['aggregation_rules']['edition'] === 'standard');
        });
        $offeringItemsMap = Arr::map($applicationCounters, 'counter_name', 'offering_item_name');
        $stdCountersMap = [];
        $aliases = [
            'PartnerHostedStorage' => 'ServiceProviderHostedStorage',
            'DrPartnerHostedStorage' => 'DRPartnerHostedStorage',
            'DrAcronisHostedStorage' => 'DRAcronisHostedStorage',
            'PublicIPsCount' => 'PublicIPsCount',
            'PublicIpsCount' => 'PublicIpsCount',
            'O365SeatsCount' => 'O365SeatsCount', // For standard, "MailboxCount" is used instead (legacy)
        ];
        foreach ($stdCounters as $counter) {
            $stdCountersMap[$counter['counter_name']] = true;
        }

        $countersPrefixes = static::PREFIXES_COUNTERS;
        $allCounters = [];
        foreach ($applicationCounters as $counter) {
            $counterName = $counter['counter_name'];
            $counterIsForUser = Arr::get($counter, 'is_for_user', false);
            foreach ($countersPrefixes as $prefix) {
                $possibleOriginal = mb_substr($counterName, mb_strlen($prefix));
                if (strpos($counterName,
                        $prefix) === 0 && (isset($stdCountersMap[$possibleOriginal]) || isset($aliases[$possibleOriginal]))) {
                    if (isset($aliases[$possibleOriginal])) {
                        $possibleOriginal = $aliases[$possibleOriginal];
                    }

                    $allCounters[$possibleOriginal] = Arr::get($allCounters, $possibleOriginal, []);
                    $allCounters[$possibleOriginal][] = [
                        'name' => $counterName,
                        'isForUser' => $counterIsForUser,
                    ];
                    break;
                }
            }
        }

        // Hard code missed counters
        $allCounters['LocalStorageUsage'] = [];
        $allCounters['MailboxCount'] = [];
        $offeringItemsMap['LocalStorageUsage'] = 'local_storage';
        $offeringItemsMap['MailboxCount'] = 'mailboxes';

        return [
            'counters' => $allCounters,
            'offering_items' => $offeringItemsMap,
        ];
    }

    public function buildOfferingItemNameNormalizationMap()
    {
        $normalizationMap = static::NORMALIZATION_MAP;
        $offeringItemsPrefixes = static::PREFIXES_OFFERING_ITEMS;
        foreach ($this->getAllCounters() as $counter) {
            $offeringItemName = $stdOfferingItemName = $counter->getOfferingItemName();

            if (isset($normalizationMap[$offeringItemName])) {
                continue;
            }

            foreach ($offeringItemsPrefixes as $prefix) {
                if (strpos($offeringItemName, $prefix) === 0) {
                    $stdOfferingItemName = mb_substr($offeringItemName, mb_strlen($prefix));
                    break;
                }
            }

            $normalizationMap[$offeringItemName] = $stdOfferingItemName;
        }

        return $normalizationMap;
    }

    public function convertValueToOfferingItemUnits(Counter $counter, $counterValue)
    {
        if ($counterValue === 0) {
            return 0;
        }

        if ($counterValue === Counter::UNLIMITED_VALUE) {
            return ApiInterface::UNLIMITED_OFFERING_ITEM_VALUE;
        }

        return Uom::convert($counterValue, $counter->getUnit(), $counter->getOfferingItemUnit());
    }

    public function convertValueFromOfferingItemUnits(Counter $counter, $offeringItemValue)
    {
        if ($offeringItemValue === 0) {
            return 0;
        }

        if ($offeringItemValue === ApiInterface::UNLIMITED_OFFERING_ITEM_VALUE) {
            return Counter::UNLIMITED_VALUE;
        }

        return Uom::convert($offeringItemValue, $counter->getOfferingItemUnit(), $counter->getUnit());
    }
}