<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\UsageReport\Aggregation;

use Acronis\App\CloudApi\Api;
use Acronis\App\Services\Locator;
use Acronis\App\Services\Logger\AppLoggerAdapter;
use Acronis\App\Services\Logger\AppLoggerFactory;
use Acronis\App\UsageReport\CsvFormatParser;
use Acronis\App\UsageReport\ReportRowWrapperInterface;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\Cloud\Client\Model\Infra\Infra;
use Service_BaseClient_HTTPException;
use Service_MspApiClient_MspApiFactoryTrait;
use Service_MspApiClient_Api;
use Str;

class AffinityHelper
{
    use Service_MspApiClient_MspApiFactoryTrait;
    use MemoizeTrait;

    /** @var Api */
    private $cloudApi;

    /** @var \Service_MspApiClient_Api */
    private $mspApi;

    /** @var string */
    private $apsInstanceId;

    /**
     * InheritedHelper constructor.
     * @param Api $cloudApi
     */
    public function __construct($cloudApi, $apsInstanceId)
    {
        $this->cloudApi = $cloudApi;
        $this->apsInstanceId = $apsInstanceId;
    }

    /**
     * @param ReportRowWrapperInterface $reportRowWrapper
     * @throws \Exception
     * @return bool
     */
    public function resolveAffinity($reportRowWrapper)
    {
        $infraOwnerId = $reportRowWrapper->getInfraOwnerId();
        if ($infraOwnerId === $reportRowWrapper->getTenantId()) {
            return CsvFormatParser::AFFINITY_OWN;
        }

        $me = $this->getCloudApi()->getMe();
        $rootTenantId = $me->getTenantId();

        if ($infraOwnerId === $rootTenantId) {
            // Backend type is specified for Acronis storages only, so all of them are inherited for partners
            return CsvFormatParser::AFFINITY_INHERITED;
        }

        $isInherited = $this->isInfraInherited($reportRowWrapper->getTenantId(), $reportRowWrapper->getInfraId());

        return $isInherited ? CsvFormatParser::AFFINITY_INHERITED : CsvFormatParser::AFFINITY_CHILD;
    }

    protected function getCloudApi()
    {
        return $this->cloudApi;
    }

    /**
     * @throws \Exception
     * @return Service_MspApiClient_Api
     */
    protected function getMspApi()
    {
        if (is_null($this->mspApi)) {
            $url = $this->getCloudApi()->getUrl();
            $login = $this->getCloudApi()->getLogin();
            $password = $this->getCloudApi()->getPassword();
            $instanceId = $this->apsInstanceId;

            $this->mspApi = $this->createMspApiInstance($url, $login, $password, $instanceId);
        }

        return $this->mspApi;
    }

    /**
     * @param $tenantId
     * @return array
     * @throws \Exception
     */
    protected function fetchTenantInfras($tenantId)
    {
        return $this->memoize(function () use ($tenantId) {
            try {
                $cloudApi = $this->getCloudApi();
                $uuids = $cloudApi->getTenantInfrasUuids($tenantId);
                $infras = $cloudApi->fetchInfras($uuids);

                return $infras;
            } catch (Service_BaseClient_HTTPException $e) {
                $errorMessage = Str::format(
                    'Cannot find infras for tenant "%s". Error: "%s".',
                    $tenantId, $e->getMessage()
                );

                /** @var AppLoggerAdapter $log */
                $log = Locator::getInstance()->get(AppLoggerFactory::NAME);
                $log->error($errorMessage);

                throw new \RuntimeException($errorMessage, $e->getCode(), $e);
            }
        }, $tenantId);
    }

    /**
     * @param string $tenantId
     * @param string $infraId
     * @throws \Exception
     * @return bool
     */
    private function isInfraInherited($tenantId, $infraId)
    {
        $infras = $this->fetchTenantInfras($tenantId);
        // In fact, getStorages returns list of inherited infras and own storages. If it will occasionally return
        // child storages we will filter them by inherited = 0. Own affinity is returned before this method call, so we don't hane to
        // distinguish child and own storages.
        // If storage is not present in this list or it is not inherited it could be only a child storage,
        // if present and marked inherited it is inherited.
        $matchingInfras = array_filter($infras, function ($infra) use ($infraId) {
            /** @var Infra $infra */
            return $infra->getId() === $infraId;
        });

        return count($matchingInfras) > 0;
    }
}