<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

use Acronis\App\CloudApi\Api;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Services\OssApi\OssApiAwareTrait;

abstract class APSResource extends \APS\ResourceBase
{
    use OssApiAwareTrait;
    use CountersManagerAwareTrait;

    const PACKAGE_ID = 'http://www.acronis.com/aps/cloud-backup-v1';
    const CLOUD_BACKUP_GLOBALS_TYPE = 'http://acronis.com/aps/cloud-backup-v1/globals/9.6';
    const CLOUD_BACKUP_RESELLER_TYPE = 'http://acronis.com/aps/cloud-backup-v1/reseller/1.0';
    const CLOUD_BACKUP_TENANT_TYPE = "http://acronis.com/aps/cloud-backup-v1/tenant/9.8";
    const CLOUD_BACKUP_USERS_TYPE = "http://acronis.com/aps/cloud-backup-v1/users/7.9";
    const CLOUD_BACKUP_PROFILE_TYPE = 'http://acronis.com/aps/cloud-backup-v1/backup-profile/5.1';
    const CLOUD_BACKUP_DATACENTER_TYPE = 'http://acronis.com/aps/cloud-backup-v1/datacenter/4.1';
    const CLOUD_BACKUP_DATACENTER_TYPE_MAJOR = 'http://acronis.com/aps/cloud-backup-v1/datacenter/4.0';
    const CLOUD_BACKUP_MINIMAL_COMMITMENT_TIER = 'http://acronis.com/aps/cloud-backup-v1/minimal-commitment-tier/4.0';

    const EXCEPTION_CAN_NOT_FIND_DATACENTER_MESSAGE = 'Can\'t find Datacenter by ID "%s"';

    const ERROR_RESOURCES_NOT_READY = 'Resource(s) not ready. Finish operations with all resource(s) first.';
    const ERROR_RESOURCE_NOT_READY = ' Resource type: %s, resource id: %s, resource status: %s.';

    const UNDERSCORE_DELIMITER = '%s_%s';

    const EXCEPTION_METHOD_IS_NOT_APPLICABLE_MESSAGE = 'Method "%s" is not applicable.';
    const BASH_APS_PHP_RUNTIME_VERSION = 'rpm -qa | grep aps-php-runtime |  awk -F"-" \'{print $4}\'';

    const ASYNC_OPERATION_TIMEOUT = 1;

    const LOG_CALLED_METHOD_WAS_CALLED = 'Method "%s" was called';

    const PACKAGE_TYPE_PREMIUM = 'premium';

    /**
     * @var FacadeInterface
     */
    protected $_facade;

    /**
     * @var Model_ManagerFactoryInterface
     */
    protected $_managerFactory;

    /**
     * @return Session_StoreInterface
     */
    protected $_sessionStore;

    /**
     * @var \APS\ControllerProxy
     */
    protected $_apsc;

    /** @var Service_MspApiClient_ApiInterface */
    private $_mspApi;

    /** @var  Service_APSApi_ApiInterface */
    private $_apsApi;

    /** @var  Service_PbaApi_ApiInterface */
    private $_pbaApi;

    /** @var BuildInfoInterface */
    private $_buildInfo;

    /**
     * @param FacadeInterface $facade
     * @param Model_ManagerFactoryInterface $managerFactory
     * @param Session_StoreInterface $sessionStore
     * @param APS\ControllerProxy $apsc
     * @param BuildInfoInterface $buildInfo
     * @throws \Exception
     */
    public function __construct(
        FacadeInterface $facade = NULL,
        Model_ManagerFactoryInterface $managerFactory = NULL,
        Session_StoreInterface $sessionStore = NULL,
        \APS\ControllerProxy $apsc = NULL,
        BuildInfoInterface $buildInfo = NULL
    )
    {
        parent::__construct();

        $this->_facade = is_null($facade) ? new Facade() : $facade;
        $this->_managerFactory = is_null($managerFactory) ? new Model_ManagerFactory() : $managerFactory;
        $this->_apsc = is_null($apsc) ? \APS\Request::getController() : $apsc;
        $this->_buildInfo = $buildInfo;

        if (is_null($sessionStore))
        {
            $sessionStore = Session_Store::getInstance();
            $sessionStore->setSessionId($this->_apsc->getInstanceId());
        }
        $this->_sessionStore = $sessionStore;
    }

    public function update()
    {
        $this->_apsc->updateResource($this);
    }

    /**
     * @return PartnerCredentials
     */
    abstract public function getPartnerCredentials();

    public function getApsRuntimePackageVersion()
    {
        //ToDo: move it into endpoint.sh for writing to file and read version from file
        return trim(shell_exec(self::BASH_APS_PHP_RUNTIME_VERSION));
    }

    /**
     * @param bool $force
     * @return Service_MspApiClient_Api
     */
    public function getMspApi($force = false)
    {
        if (is_null($this->_mspApi))
        {
            $partnerCredentials = $this->getPartnerCredentials();
            $instanceId = sprintf(
                self::UNDERSCORE_DELIMITER,
                $this->_apsc->getInstanceId(),
                md5($partnerCredentials->partnerLogin . $partnerCredentials->baasUrl)
            );
            $api = $this->_facade->getMspApi($partnerCredentials->baasUrl, $instanceId);
            $api->setCredentials(
                $partnerCredentials->partnerLogin,
                $partnerCredentials->partnerPassword,
                $partnerCredentials->baasUrl,
                $force
            );
            if ($force)
            {
                $api->login();
            }

            $this->_mspApi = $api;
        }
        return $this->_mspApi;
    }

    /**
     * @return Service_PbaApi_Api
     * @throws Exception
     */
    public function _getPbaApi()
    {
        if (is_null($this->_pbaApi))
        {
            $config = Settings::getInstance()->getBillingConfig();
            $api = $this->_facade->getPbaApi($config);
            $this->_pbaApi = $api;
        }
        return $this->_pbaApi;
    }

    /**
     * @return Service_APSApi_Api
     */
    public function _getApsApi()
    {
        if (is_null($this->_apsApi))
        {
            $this->_apsApi = $this->_facade->getApsApi($this->_apsc);
        }
        return $this->_apsApi;
    }

    public function _getOpenApi()
    {
        return $this->getOssApi();
    }

    /**
     * @param string $key
     * @return string
     */
    public function getCacheUniqueKey($key)
    {
        return sprintf(self::UNDERSCORE_DELIMITER, $this->aps->id, $key);
    }

    /**
     * Storing data in the cache
     *
     * @param string $key
     * @return mixed
     */
    public function getValueFromCache($key)
    {
        $uniqueKey = $this->getCacheUniqueKey($key);
        $this->_sessionStore->startSession();
        $data = $this->_sessionStore->getValue($uniqueKey);
        $this->_sessionStore->stopSession();
        return $data;
    }

    /**
     * Restoring data from the cache
     *
     * @param string $key
     * @param mixed $data
     * @param int|null $ttl
     */
    public function setValueToCache($key, $data, $ttl=NULL)
    {
        $uniqueKey = $this->getCacheUniqueKey($key);
        $this->_sessionStore->startSession();
        $this->_sessionStore->setValue($uniqueKey, $data, $ttl);
        $this->_sessionStore->stopSession();
    }

    /**
     * Delete data in the session
     *
     * @param string $key
     * @return mixed
     */
    public function unsetValueFromCache($key)
    {
        $uniqueKey = $this->getCacheUniqueKey($key);
        $this->_sessionStore->startSession();
        $data = $this->_sessionStore->unsetValue($uniqueKey);
        $this->_sessionStore->stopSession();
        return $data;
    }

    /**
     * @param $datacenter
     * @return ApiInterface
     */
    protected function createCloudApiInstance($datacenter)
    {
        return new Api(
            $datacenter->url,
            $datacenter->login,
            $datacenter->password
        );
    }

    /**
     * @return BuildInfoInterface
     */
    protected function _getBuildInfo()
    {
        if (is_null($this->_buildInfo))
        {
            $this->_buildInfo = new BuildInfo();
        }

        return $this->_buildInfo;
    }

    /**
     * @return string|NULL
     */
    protected static function _getApsIdentityId()
    {
        return isset($_SERVER['HTTP_APS_IDENTITY_ID']) ? $_SERVER['HTTP_APS_IDENTITY_ID'] : NULL;
    }

    /**
     * @param globals $global
     * @param string $id
     * @return object
     * @throws Exception
     */
    protected function _getDatacenterById($global, $id)
    {
        $datacenters = $global->datacenters;
        foreach ($datacenters as $datacenter)
        {
            if ($datacenter->id == $id)
            {
                return $datacenter;
            }
        }
        $message = sprintf(self::EXCEPTION_CAN_NOT_FIND_DATACENTER_MESSAGE, $id);
        Log::getInstance()->error($message);
        throw new Exception($message);
    }

    /**
     * @param \APS\EventSubscription $event
     * @param string $resourceId
     */
    protected function subscribeOnEventById($event, $resourceId)
    {
        if (!is_object($event->source)) {
            $event->source = new stdClass();
        }
        $event->source->id = $resourceId;
        $this->_apsc->impersonate($this)->subscribe($this, $event);
    }
    /**
     * @param string $eventType
     * @param string $resourceId
     */
    protected function unsubscribeFromEventById($eventType, $resourceId)
    {
        $this->unsubscribeFromEvent(function(\APS\EventSubscription $subscription) use ($eventType, $resourceId) {
            return $subscription->event === $eventType && isset($subscription->source->id) && $subscription->source->id === $resourceId;
        });
    }

    /**
     * @param \APS\EventSubscription $event
     * @param string $resourceType
     */
    protected function subscribeOnEventByType($event, $resourceType)
    {
        if (!is_object($event->source)) {
            $event->source = new stdClass();
        }
        $event->source->type = $resourceType;
        $this->_apsc->impersonate($this)->subscribe($this, $event);
    }

    /**
     * @param string $eventType
     * @param string $resourceType
     */
    protected function unsubscribeFromEventByType($eventType, $resourceType)
    {
        $this->unsubscribeFromEvent(function(\APS\EventSubscription $subscription) use ($eventType, $resourceType) {
            return $subscription->event === $eventType && isset($subscription->source->type) && $subscription->source->type === $resourceType;
        });
    }

    /**
     * @param callable $checkEventSubscription
     */
    private function unsubscribeFromEvent($checkEventSubscription)
    {
        $apsc = $this->_apsc->impersonate($this);
        $subscriptions = $apsc->subscriptions($this);
        if (empty($subscriptions)) {
            return;
        }
        foreach ($subscriptions as $subscription) {
            if (!$checkEventSubscription($subscription)) {
                continue;
            }
            $apsc->unsubscribe($this, $subscription);
        }
    }

    protected function getAllCounters()
    {
        return $this->getCountersManager()->getAllCountersMap();
    }

}
