<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

// todo move to namespace Utils
class Arr
{
    const VALUE_DELIMITER = '=';
    const PAIR_DELIMITER = ';';

    /**
     * @param array $values
     * @return string
     */
    public static function encode(array $values)
    {
        $pairs = [];
        foreach ($values as $k => $v) {
            $pairs[] = is_null($v) ? $k : $k . self::VALUE_DELIMITER . $v;
        }

        return implode(self::PAIR_DELIMITER, $pairs);
    }

    /**
     * @param string $string
     * @return array
     */
    public static function decode($string)
    {
        if (!is_string($string) || $string === '') {
            return [];
        }

        $pairs = explode(self::PAIR_DELIMITER, $string);

        $values = [];
        foreach ($pairs as $pair) {
            $parts = explode(self::VALUE_DELIMITER, $pair, 2);
            $values[$parts[0]] = count($parts) > 1 ? $parts[1] : null;
        }

        return $values;
    }

    /**
     * @param array|ArrayAccess|object $array
     * @param string $key
     * @return bool
     */
    public static function has($array, $key)
    {
        if (is_array($array)) {
            return array_key_exists($key, $array);
        }

        if (is_object($array)) {
            if ($array instanceof ArrayAccess) {
                return $array->offsetExists($key);
            } else {
                // First part catches the case when a property exists but a value equals NULL
                // Second part catches the case when a property does not exist but there is magic method __isset
                return array_key_exists($key, get_object_vars($array))
                    || isset($array->{$key});
            }
        }

        return false;
    }

    /**
     * @param array|ArrayAccess|object $array
     * @param array[] $keys
     * @return bool
     */
    public static function hasNestedItem($array, $keys)
    {
        $currentKey = reset($keys);
        $hasCurrentKey = static::has($array, $currentKey);

        if (count($keys) == 1) {
            return $hasCurrentKey;
        }

        return $hasCurrentKey && static::hasNestedItem($array, array_slice($keys, 1));
    }

    /**
     * @param array|ArrayAccess|object $array
     * @param string[] $keys
     * @param mixed $defaultValue
     * @return mixed
     */
    public static function getNestedItem($array, $keys, $defaultValue)
    {
        $currentKey = reset($keys);
        $currentValue = static::get($array, $currentKey, $defaultValue);

        if (count($keys) == 1) {
            return $currentValue;
        }

        return static::getNestedItem($currentValue, array_slice($keys, 1), $defaultValue);
    }

    /**
     * @param array|ArrayAccess|object $array
     * @param string $key
     * @param mixed $defaultValue
     * @return mixed
     */
    public static function get($array, $key, $defaultValue = null)
    {
        if (!static::has($array, $key)) {
            return $defaultValue;
        }

        if (is_array($array)) {
            return $array[$key];
        }

        if (is_object($array)) {
            if ($array instanceof ArrayAccess) {
                return $array->offsetGet($key);
            } else {
                return $array->{$key};
            }
        }

        return $defaultValue;
    }

    /**
     * @param array|Iterator $array
     * @param mixed|callable $value
     * @param mixed $defaultValue
     * @return mixed
     */
    public static function find($array, $value, $defaultValue = null)
    {
        $isValueCallable = is_callable($value);
        foreach ($array as $item) {
            if ($isValueCallable && $value($item)) {
                return $item;
            } elseif ($item === $value) {
                return $item;
            }
        }

        return $defaultValue;
    }

    /**
     * @param string $key
     * @param array|ArrayAccess|object $array
     * @param int $value
     * @return mixed
     */
    public static function set(&$array, $key, $value)
    {
        if (is_array($array)) {
            $array[$key] = $value;
        }

        if (is_object($array)) {
            if ($array instanceof ArrayAccess) {
                $array->offsetSet($key, $value);
            } else {
                $array->{$key} = $value;
            }
        }
    }

    /**
     * Gets array/object key, 0 by default, increments by value, sets key.
     * @param array $array
     * @param string $key
     * @param int $value
     */
    public static function increment(&$array, $key, $value)
    {
        $increasedValue = static::get($array, $key, 0) + $value;
        static::set($array, $key, $increasedValue);
    }

    /**
     * Example:
     *     $limits = ['a' => 11, 'b' => 22, 'c' => 33, 'd' => 44,];
     *     $usages = ['a' => 1,  'b' => 2,  'c' => 3,  'e' => 4,];
     *     $counters = Arr::merge([
     *          'limit' => $limits,
     *          'usage' => $usages,
     *     ], [
     *          'usage' => 0,
     *     ]);
     * Variable '$counters' equals to
     * [
     *     'a' => ['limit' => 11, 'usage' => 1,],
     *     'b' => ['limit' => 22, 'usage' => 2,],
     *     'c' => ['limit' => 33, 'usage' => 3,],
     *     'd' => ['limit' => 44, 'usage' => 0,],
     *     'e' => ['usage' => 4,],
     * ]
     *
     * @param array $arrays
     * @param array $defaultValues
     * @return array
     */
    public static function merge(array $arrays, array $defaultValues = [])
    {
        $rootKeys = array_keys(array_merge(...array_values($arrays)));

        $results = [];
        foreach ($rootKeys as $rootKey) {
            $values = [];
            foreach ($arrays as $key => $array) {
                if (static::has($array, $rootKey)) {
                    $values[$key] = static::get($array, $rootKey);
                    continue;
                }

                if (static::has($defaultValues, $key)) {
                    $values[$key] = static::get($defaultValues, $key);
                    continue;
                }
                // otherwise do not create the key in values
            }

            $results[$rootKey] = $values;
        }

        return $results;
    }

    /**
     * Example:
     *     $countries = [
     *        ['country_code' => 'US', 'currency_code' => 'USD', 'name' => 'United States',],
     *        ['country_code' => 'RU', 'currency_code' => 'RUB', 'name' => 'Russia',],
     *        ['country_code' => 'FR', 'currency_code' => 'EUR', 'name' => 'France',],
     *     ];
     *     $currencies = Arr::map($countries, 'country_code', 'currency_code')
     * Variable '$currencies' equals to
     * [
     *     'US' => 'USD',
     *     'RU' => 'RUB',
     *     'FR' => 'EUR',
     * ]
     *
     * @param array $items
     * @param callable|string $keyProperty
     * @param callable|string $valueProperty
     * @param bool $unique true - throw an exception if the key is not unique
     * @return array
     */
    public static function map(array $items, $keyProperty, $valueProperty = null, $unique = true)
    {
        $map = [];
        foreach ($items as $itemKey => $item) {
            $key = static::retrieveProperty($keyProperty, $item, $itemKey);
            $value = $valueProperty === null
                ? $item
                : static::retrieveProperty($valueProperty, $item, $itemKey);

            if ($unique && static::has($map, $key)) {
                throw new InvalidArgumentException(Str::format(
                    'Key "%s" is not unique.',
                    $key
                ));
            }

            $map[$key] = $value;
        }

        return $map;
    }

    /**
     * @param $arr1
     * @param $arr2
     * @return bool
     */
    public static function equalValues($arr1, $arr2)
    {
        sort($arr1);
        sort($arr2);

        return $arr1 == $arr2;
    }

    private static function retrieveProperty($property, $item, $key)
    {
        if (is_callable($property)) {
            return $property($item, $key);
        }

        return static::get($item, $property, null);
    }
}
